# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.cp import IntervalVar, Step, Pulse
from pyomo.contrib.cp.scheduling_expr.step_function_expressions import (
    AlwaysIn,
    CumulativeFunction,
    NegatedStepFunction,
    StepAtStart,
    StepAtEnd,
    StepAt,
)

from pyomo.environ import ConcreteModel, LogicalConstraint


class CommonTests(unittest.TestCase):
    def get_model(self):
        m = ConcreteModel()
        m.a = IntervalVar()
        m.b = IntervalVar()
        m.c = IntervalVar([1, 2])

        return m


class TestPulse(CommonTests):
    def test_bad_interval_var(self):
        with self.assertRaisesRegex(
            TypeError,
            "The 'interval_var' argument for a 'Pulse' must "
            "be an 'IntervalVar'.\n"
            "Received: <class 'float'>",
        ):
            thing = Pulse(interval_var=1.2, height=4)

    def test_create_pulse_with_scalar_interval_var(self):
        m = self.get_model()
        p = Pulse(interval_var=m.a, height=1)

        self.assertIsInstance(p, Pulse)
        self.assertEqual(str(p), "Pulse(a, height=1)")

    def test_create_pulse_with_interval_var_data(self):
        m = self.get_model()
        p = Pulse(interval_var=m.c[2], height=2)
        self.assertIsInstance(p, Pulse)
        self.assertEqual(str(p), "Pulse(c[2], height=2)")


class TestStep(CommonTests):
    def test_bad_time_point(self):
        m = self.get_model()
        with self.assertRaisesRegex(
            TypeError,
            "The 'time' argument for a 'Step' must be either "
            r"an 'IntervalVarTimePoint' \(for example, the "
            r"'start_time' or 'end_time' of an IntervalVar\) or "
            "an integer time point in the time horizon.\n"
            "Received: "
            "<class 'pyomo.contrib.cp.interval_var.ScalarIntervalVar'>",
        ):
            thing = Step(m.a, height=2)


class TestSumStepFunctions(CommonTests):
    def test_sum_step_and_pulse(self):
        m = self.get_model()
        expr = Step(m.a.start_time, height=4) + Pulse((m.b, -1))

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 2)
        self.assertEqual(len(expr.args), 2)
        self.assertIsInstance(expr.args[0], StepAtStart)
        self.assertIsInstance(expr.args[1], Pulse)

        self.assertEqual(
            str(expr), "Step(a.start_time, height=4) + Pulse(b, height=-1)"
        )

    def test_args_clone_correctly(self):
        m = self.get_model()
        expr = Step(m.a.start_time, height=4) + Pulse((m.b, -1))
        expr2 = expr + Step(m.b.end_time, height=4)

        self.assertIsInstance(expr2, CumulativeFunction)
        self.assertEqual(len(expr2.args), 3)
        self.assertEqual(expr2.nargs(), 3)
        self.assertIsInstance(expr2.args[0], StepAtStart)
        self.assertIsInstance(expr2.args[1], Pulse)
        self.assertIsInstance(expr2.args[2], StepAtEnd)

        # This will force expr to clone its arguments because it did the
        # appending trick to make expr2.
        expr3 = expr + Pulse(interval_var=m.b, height=-5)

        self.assertIsInstance(expr3, CumulativeFunction)
        self.assertEqual(len(expr3.args), 3)
        self.assertEqual(expr3.nargs(), 3)
        self.assertIsInstance(expr3.args[0], StepAtStart)
        self.assertIsInstance(expr3.args[1], Pulse)
        self.assertIsInstance(expr3.args[2], Pulse)

    def test_args_clone_correctly_in_place(self):
        m = self.get_model()
        s1 = Step(m.a.start_time, height=1)
        s2 = Step(m.b.end_time, height=1)
        s3 = Step(m.b.start_time, height=2)
        p = Pulse(interval_var=m.b, height=3)

        e1 = s1 + s2
        e2 = e1 + s3
        e3 = e1
        e3 += p

        self.assertIsInstance(e1, CumulativeFunction)
        self.assertEqual(e1.nargs(), 2)
        self.assertIs(e1.args[0], s1)
        self.assertIs(e1.args[1], s2)

        self.assertIsInstance(e2, CumulativeFunction)
        self.assertEqual(e2.nargs(), 3)
        self.assertIs(e2.args[0], s1)
        self.assertIs(e2.args[1], s2)
        self.assertIs(e2.args[2], s3)

        self.assertIsInstance(e3, CumulativeFunction)
        self.assertEqual(e3.nargs(), 3)
        self.assertIs(e3.args[0], s1)
        self.assertIs(e3.args[1], s2)
        self.assertIs(e3.args[2], p)

    def test_sum_two_pulses(self):
        m = self.get_model()
        m.p1 = Pulse(interval_var=m.a, height=3)
        m.p2 = Pulse(interval_var=m.b, height=-2)

        expr = m.p1 + m.p2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)
        self.assertIs(expr.args[0], m.p1)
        self.assertIs(expr.args[1], m.p2)

    def test_sum_in_place(self):
        m = self.get_model()
        expr = Step(m.a.start_time, height=4) + Pulse(interval_var=m.b, height=-1)
        expr += Step(0, 1)

        self.assertEqual(len(expr.args), 3)
        self.assertEqual(expr.nargs(), 3)
        self.assertIsInstance(expr.args[0], StepAtStart)
        self.assertIsInstance(expr.args[1], Pulse)
        self.assertIsInstance(expr.args[2], StepAt)

        self.assertEqual(
            str(expr),
            "Step(a.start_time, height=4) + Pulse(b, height=-1) + Step(0, height=1)",
        )

    def test_sum_steps_in_place(self):
        m = self.get_model()
        s1 = Step(m.a.end_time, height=2)
        expr = s1

        # Just a step function
        self.assertIsInstance(expr, StepAtEnd)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)

        s2 = Step(m.b.end_time, height=3)
        expr += s2

        # becomes a cumulative function
        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)
        self.assertIs(expr.args[0], s1)
        self.assertIs(expr.args[1], s2)

    def test_sum_pulses_in_place(self):
        m = self.get_model()
        p1 = Pulse(interval_var=m.a, height=2)
        expr = p1

        self.assertIsInstance(expr, Pulse)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)

        p2 = Pulse(interval_var=m.b, height=3)
        expr += p2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)
        self.assertIs(expr.args[0], p1)
        self.assertIs(expr.args[1], p2)

    def test_sum_step_and_cumul_func(self):
        m = self.get_model()
        s1 = Step(m.a.start_time, height=4)
        p1 = Step(m.a.start_time, height=4)
        cumul = s1 + p1
        s = Step(m.a.end_time, height=3)
        expr = s + cumul

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 3)
        self.assertIs(expr.args[0], s)
        self.assertIs(expr.args[1], s1)
        self.assertIs(expr.args[2], p1)

    def test_subtract_cumul_from_pulse(self):
        m = self.get_model()
        p1 = Pulse(interval_var=m.a, height=2)
        s1 = Step(m.a.start_time, height=4)
        p2 = Pulse(interval_var=m.b, height=3)
        cumul = s1 - p2
        expr = p1 - cumul

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 3)
        self.assertIs(expr.args[0], p1)
        self.assertIsInstance(expr.args[1], NegatedStepFunction)
        self.assertIs(expr.args[1].args[0], s1)
        self.assertIsInstance(expr.args[2], NegatedStepFunction)
        self.assertIsInstance(expr.args[2].args[0], NegatedStepFunction)
        self.assertIs(expr.args[2].args[0].args[0], p2)

    def test_subtract_two_cumul_functions(self):
        m = self.get_model()
        p1 = Pulse(interval_var=m.a, height=2)
        s1 = Step(m.a.start_time, height=4)
        p2 = Pulse(interval_var=m.b, height=3)
        p3 = Pulse(interval_var=m.a, height=-4)
        cumul1 = s1 - p2
        cumul2 = p2 + p3
        expr = cumul1 - cumul2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 4)
        self.assertIs(expr.args[0], s1)
        self.assertIsInstance(expr.args[1], NegatedStepFunction)
        self.assertIs(expr.args[1].args[0], p2)
        self.assertIsInstance(expr.args[2], NegatedStepFunction)
        self.assertIs(expr.args[2].args[0], p2)
        self.assertIsInstance(expr.args[3], NegatedStepFunction)
        self.assertIs(expr.args[3].args[0], p3)

    def test_subtract_two_cumul_functions_requiring_cloning(self):
        m = self.get_model()
        p1 = Pulse(interval_var=m.a, height=2)
        s1 = Step(m.a.start_time, height=4)
        p2 = Pulse(interval_var=m.b, height=3)
        p3 = Pulse(interval_var=m.a, height=-4)
        cumul1 = s1 - p2
        # This will append to the args of cumul1, and then we'll have to clone
        # them when we make expr
        aux = cumul1 + Step(0, 4)
        cumul2 = p2 + p3
        expr = cumul1 - cumul2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 4)
        self.assertIs(expr.args[0], s1)
        self.assertIsInstance(expr.args[1], NegatedStepFunction)
        self.assertIs(expr.args[1].args[0], p2)
        self.assertIsInstance(expr.args[2], NegatedStepFunction)
        self.assertIs(expr.args[2].args[0], p2)
        self.assertIsInstance(expr.args[3], NegatedStepFunction)
        self.assertIs(expr.args[3].args[0], p3)

    def test_sum_two_cumul_funcs(self):
        m = self.get_model()
        s1 = Step(m.a.start_time, height=4)
        p1 = Step(m.a.start_time, height=4)
        cumul1 = s1 + p1
        s2 = Step(m.a.end_time, height=3)
        s3 = Step(0, height=34)
        cumul2 = s2 + s3
        expr = cumul1 + cumul2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 4)
        self.assertIs(expr.args[0], s1)
        self.assertIs(expr.args[1], p1)
        self.assertIs(expr.args[2], s2)
        self.assertIs(expr.args[3], s3)

    def test_sum_two_cumul_funcs_requiring_cloning_args(self):
        m = self.get_model()
        s1 = Step(m.a.start_time, height=4)
        p1 = Step(m.a.start_time, height=4)
        cumul1 = s1 + p1
        # This one will extend cumul1, so we'll have to clone it when we build
        # expr
        aux = cumul1 + Step(5, 4)

        s2 = Step(m.a.end_time, height=3)
        s3 = Step(0, height=34)
        cumul2 = s2 + s3
        expr = cumul1 + cumul2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 4)
        self.assertIs(expr.args[0], s1)
        self.assertIs(expr.args[1], p1)
        self.assertIs(expr.args[2], s2)
        self.assertIs(expr.args[3], s3)

    def test_cannot_add_constant(self):
        m = self.get_model()
        with self.assertRaisesRegex(
            TypeError,
            "Cannot add object of class <class 'int'> to object of class "
            "<class 'pyomo.contrib.cp.scheduling_expr."
            "step_function_expressions.StepAtStart'>",
        ):
            expr = Step(m.a.start_time, height=6) + 3

    def test_cannot_add_to_constant(self):
        m = self.get_model()
        with self.assertRaisesRegex(
            TypeError,
            "Cannot add object of class <class 'pyomo.contrib.cp."
            "scheduling_expr.step_function_expressions.StepAtStart'> to "
            "object of class <class 'int'>",
        ):
            expr = 4 + Step(m.a.start_time, height=6)

    def test_python_sum_funct(self):
        # We allow adding to 0 so that sum() works as expected
        m = self.get_model()
        expr = sum(Pulse(interval_var=m.c[i], height=1) for i in [1, 2])

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)
        self.assertIsInstance(expr.args[0], Pulse)
        self.assertIsInstance(expr.args[1], Pulse)


class TestSubtractStepFunctions(CommonTests):
    def test_subtract_two_steps(self):
        m = self.get_model()

        s = Step(m.a.start_time, height=2) - Step(m.b.start_time, height=5)

        self.assertIsInstance(s, CumulativeFunction)
        self.assertEqual(len(s.args), 2)
        self.assertEqual(s.nargs(), 2)
        self.assertIsInstance(s.args[0], StepAtStart)
        self.assertIsInstance(s.args[1], NegatedStepFunction)
        self.assertEqual(len(s.args[1].args), 1)
        self.assertEqual(s.args[1].nargs(), 1)
        self.assertIsInstance(s.args[1].args[0], StepAtStart)

    def test_subtract_step_and_pulse(self):
        m = self.get_model()
        s1 = Step(m.a.end_time, height=2)
        s2 = Step(m.b.start_time, height=5)
        p = Pulse(interval_var=m.a, height=3)

        expr = s1 - s2 - p

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 3)
        self.assertEqual(expr.nargs(), 3)
        self.assertIs(expr.args[0], s1)
        self.assertIsInstance(expr.args[1], NegatedStepFunction)
        self.assertIs(expr.args[1].args[0], s2)
        self.assertIsInstance(expr.args[2], NegatedStepFunction)
        self.assertIs(expr.args[2].args[0], p)

    def test_subtract_pulse_from_two_steps(self):
        m = self.get_model()
        s1 = Step(m.a.end_time, height=2)
        s2 = Step(m.b.start_time, height=5)
        p = Pulse(interval_var=m.a, height=3)

        expr = s1 + s2 - p
        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 3)
        self.assertEqual(expr.nargs(), 3)
        self.assertIs(expr.args[0], s1)
        self.assertIs(expr.args[1], s2)
        self.assertIsInstance(expr.args[2], NegatedStepFunction)
        self.assertIs(expr.args[2].args[0], p)

    def test_args_clone_correctly(self):
        m = self.get_model()
        m.p1 = Pulse(interval_var=m.a, height=3)
        m.p2 = Pulse(interval_var=m.b, height=4)
        m.s = Step(m.a.start_time, height=-1)

        expr1 = m.p1 - m.p2
        self.assertIsInstance(expr1, CumulativeFunction)
        self.assertEqual(expr1.nargs(), 2)
        self.assertIs(expr1.args[0], m.p1)
        self.assertIsInstance(expr1.args[1], NegatedStepFunction)
        self.assertIs(expr1.args[1].args[0], m.p2)

        expr2 = m.p1 - m.s
        self.assertIsInstance(expr2, CumulativeFunction)
        self.assertEqual(expr2.nargs(), 2)
        self.assertIs(expr2.args[0], m.p1)
        self.assertIsInstance(expr2.args[1], NegatedStepFunction)
        self.assertIs(expr2.args[1].args[0], m.s)

    def test_args_clone_correctly_in_place(self):
        m = self.get_model()
        m.p1 = Pulse(interval_var=m.a, height=3)
        m.p2 = Pulse(interval_var=m.b, height=4)
        m.s = Step(m.a.start_time, height=-1)

        expr1 = m.p1 - m.p2
        # This will append p1 to expr1's args
        expr = expr1 + m.p1
        # Now we have to clone in place
        expr1 -= m.s

        self.assertIsInstance(expr1, CumulativeFunction)
        self.assertEqual(expr1.nargs(), 3)
        self.assertIs(expr1.args[0], m.p1)
        self.assertIsInstance(expr1.args[1], NegatedStepFunction)
        self.assertIs(expr1.args[1].args[0], m.p2)
        self.assertIsInstance(expr1.args[2], NegatedStepFunction)
        self.assertIs(expr1.args[2].args[0], m.s)

        # and expr is what we expect too
        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 3)
        self.assertIs(expr.args[0], m.p1)
        self.assertIsInstance(expr.args[1], NegatedStepFunction)
        self.assertIs(expr.args[1].args[0], m.p2)
        self.assertIs(expr.args[2], m.p1)

    def test_subtract_pulses_in_place(self):
        m = self.get_model()
        p1 = Pulse(interval_var=m.a, height=1)
        p2 = Pulse(interval_var=m.b, height=3)

        expr = p1
        expr -= p2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)
        self.assertIs(expr.args[0], p1)
        self.assertIsInstance(expr.args[1], NegatedStepFunction)
        self.assertIs(expr.args[1].args[0], p2)

    def test_subtract_steps_in_place(self):
        m = self.get_model()
        s1 = Step(m.a.start_time, height=1)
        s2 = Step(m.b.end_time, height=3)

        expr = s1
        expr -= s2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(len(expr.args), 2)
        self.assertEqual(expr.nargs(), 2)
        self.assertIs(expr.args[0], s1)
        self.assertIsInstance(expr.args[1], NegatedStepFunction)
        self.assertIs(expr.args[1].args[0], s2)

    def test_subtract_from_cumul_func_in_place(self):
        m = self.get_model()
        m.p1 = Pulse(interval_var=m.a, height=5)
        m.p2 = Pulse(interval_var=m.b, height=-3)
        m.s = Step(m.b.end_time, height=5)

        expr = m.p1 + m.s
        expr -= m.p2

        self.assertIsInstance(expr, CumulativeFunction)
        self.assertEqual(expr.nargs(), 3)
        self.assertIs(expr.args[0], m.p1)
        self.assertIs(expr.args[1], m.s)
        self.assertIsInstance(expr.args[2], NegatedStepFunction)
        self.assertIs(expr.args[2].args[0], m.p2)

        self.assertEqual(
            str(expr),
            "Pulse(a, height=5) + Step(b.end_time, height=5) - Pulse(b, height=-3)",
        )

    def test_cannot_subtract_constant(self):
        m = self.get_model()
        with self.assertRaisesRegex(
            TypeError,
            "Cannot subtract object of class <class 'int'> from object of "
            "class <class 'pyomo.contrib.cp."
            "scheduling_expr.step_function_expressions.StepAtStart'>",
        ):
            expr = Step(m.a.start_time, height=6) - 3

    def test_cannot_subtract_from_constant(self):
        m = self.get_model()
        with self.assertRaisesRegex(
            TypeError,
            "Cannot subtract object of class <class 'pyomo.contrib.cp."
            "scheduling_expr.step_function_expressions.StepAtStart'> from "
            "object of class <class 'int'>",
        ):
            expr = 3 - Step(m.a.start_time, height=6)


class TestAlwaysIn(CommonTests):
    def test_always_in(self):
        m = self.get_model()
        f = (
            Pulse(interval_var=m.a, height=3)
            + Step(m.b.start_time, height=2)
            - Step(m.a.end_time, height=-1)
        )

        m.cons = LogicalConstraint(expr=f.within((0, 3), (0, 10)))
        self.assertIsInstance(m.cons.expr, AlwaysIn)

        self.assertEqual(m.cons.expr.nargs(), 5)
        self.assertEqual(len(m.cons.expr.args), 5)
        self.assertIs(m.cons.expr.args[0], f)
        self.assertEqual(m.cons.expr.args[1], 0)
        self.assertEqual(m.cons.expr.args[2], 3)
        self.assertEqual(m.cons.expr.args[3], 0)
        self.assertEqual(m.cons.expr.args[4], 10)
        self.assertEqual(
            str(m.cons.expr),
            "(Pulse(a, height=3) + Step(b.start_time, height=2) - "
            "Step(a.end_time, height=-1)).within(bounds=(0, 3), "
            "times=(0, 10))",
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from weakref import ref as weakref_ref

from pyomo.common.log import is_debug_set
from pyomo.core.base.set_types import Any
from pyomo.core.base.var import IndexedVar, VarData
from pyomo.core.base.constraint import IndexedConstraint, ConstraintData
from pyomo.core.base.objective import IndexedObjective, ObjectiveData
from pyomo.core.base.expression import IndexedExpression, ExpressionData

from collections.abc import MutableMapping
from collections.abc import Mapping

logger = logging.getLogger('pyomo.core')

#
# In the future I think ComponentDict and ComponentList should inherit
# directly from (Active)IndexedComponent, and that class should be
# stripped down to a minimal interface. The abstract interface should
# be implemented on top of these classes.
#


class ComponentDict(MutableMapping):
    def __init__(self, interface_datatype, *args):
        self._interface_datatype = interface_datatype
        self._data = {}
        if len(args) > 0:
            if len(args) > 1:
                raise TypeError(
                    "ComponentDict expected at most 1 arguments, "
                    "got %s" % (len(args))
                )
            self.update(args[0])

    def construct(self, data=None):
        if is_debug_set(logger):
            logger.debug(  # pragma:nocover
                "Constructing ComponentDict object, name=%s, from data=%s"
                % (self.name, str(data))
            )
        if self._constructed:  # pragma:nocover
            return
        self._constructed = True

    #
    # Define the MutableMapping abstract methods
    #

    #
    # Currently __setitem__ only supports assignment of an explicit
    # instantiation of the interface datatype. Updates to an already
    # existing object inside this container need to be done via that
    # objects interface methods or by directly accessing object
    # attributes.
    #
    # * Note about implicit instantiation / update:
    #    Implicit instantiation or update using assignment of a valid
    #    value type could be supported, but we should more carefully
    #    consider edge cases before guaranteeing such functionality.
    #    The explicit approach provides all functionality one would
    #    need, albeit, with a slightly more verbose syntax (which one
    #    could argue is more clear).
    #
    # * Note about component ownership
    #    Currently this method only supports assignment of datatype
    #    instantiations whose parent component has not been set. This
    #    prevents the current container from "stealing" the component
    #    from a different container, leaving the old container still
    #    thinking all components it contains have their weakref
    #    pointing back to itself.
    #
    #    However, it is not clear that this would actually be an
    #    issue.  For instance, this would seem like a very effective
    #    way of "aliasing" a variable. That is, this component would
    #    only set the parent component on an incoming object
    #    assignment if that object was not already assigned a
    #    parent. All of the problem writers would not be effected by
    #    this because they generate the symbol map by hashing off of
    #    the object id(). This would also seamlessly integrate into
    #    the Suffix API as that component hashes of off id() as well.
    #
    #    Again, I would rather start with more limited functionality
    #    that we could all, without question, agree on, so this is
    #    something to consider later on.
    #
    def __setitem__(self, key, val):
        if isinstance(val, self._interface_datatype):
            if val._component is None:
                val._component = weakref_ref(self)
                if hasattr(self, "_active"):
                    self._active |= getattr(val, '_active', True)
                if key in self._data:
                    # release the current component
                    # * see __delitem__ for explanation
                    self._data[key]._component = None
                self._data[key] = val
                self._data[key]._index = key
                return
            elif (key in self._data) and (self._data[key] is val):
                # a very special case that makes sense to handle
                # because the implied order should be: (1) delete
                # the object at the current index, (2) insert the
                # the new object. This performs both without any
                # actions, but it is an extremely rare case, so
                # it should go last.
                return
            # see note about allowing components to live in more than
            # one container
            raise ValueError(
                "Invalid component object assignment to ComponentDict "
                "%s at key %s. A parent component has already been "
                "assigned the object: %s"
                % (self.name, key, val.parent_component().name)
            )
        # see note about implicit assignment and update
        raise TypeError(
            "ComponentDict must be assigned objects "
            "of type %s. Invalid type for key %s: %s"
            % (self._interface_datatype.__name__, key, type(val))
        )

    # Since we don't currently allow objects to be assigned when their
    # parent component is already set, it would make sense to reset
    # the parent component back to None for an object being removed
    # from this container; thus, allowing objects to be naturally
    # transferred between containers if one so chooses. This would be
    # more complicated if allowing components to live in more than one
    # container (see note above __setitem__), and maybe this is the
    # reason not to support that.
    def __delitem__(self, key):
        obj = self._data[key]
        obj._component = None
        del self._data[key]

    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return self._data.__len__()

    #
    # Override a few default implementations on MutableMapping
    #

    # We want to avoid generating Pyomo expressions due to
    # comparison of values; thus, we convert this to a plain
    # dictionary mapping key->(type(val), id(val)) and
    # compare that instead.
    def __eq__(self, other):
        if not isinstance(other, Mapping):
            return False
        return dict((key, (type(val), id(val))) for key, val in self.items()) == dict(
            (key, (type(val), id(val))) for key, val in other.items()
        )

    def __ne__(self, other):
        return not (self == other)


#
# ComponentDict needs to come before IndexedComponent
# (or subclasses of) so we can override certain methods
#


class VarDict(ComponentDict, IndexedVar):
    def __init__(self, *args, **kwds):
        IndexedVar.__init__(self, Any, **kwds)
        # Constructor for ComponentDict needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentDict.__init__(self, VarData, *args, **kwds)


class ConstraintDict(ComponentDict, IndexedConstraint):
    def __init__(self, *args, **kwds):
        IndexedConstraint.__init__(self, Any, **kwds)
        # Constructor for ComponentDict needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentDict.__init__(self, ConstraintData, *args, **kwds)


class ObjectiveDict(ComponentDict, IndexedObjective):
    def __init__(self, *args, **kwds):
        IndexedObjective.__init__(self, Any, **kwds)
        # Constructor for ComponentDict needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentDict.__init__(self, ObjectiveData, *args, **kwds)


class ExpressionDict(ComponentDict, IndexedExpression):
    def __init__(self, *args, **kwds):
        IndexedExpression.__init__(self, Any, **kwds)
        # Constructor for ComponentDict needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentDict.__init__(self, ExpressionData, *args, **kwds)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import Counter

from pyomo.common.dependencies import numpy as numpy, numpy_available

if numpy_available:
    from numpy.testing import assert_array_almost_equal

from pyomo.common import unittest
import pyomo.opt

from pyomo.contrib.alternative_solutions import enumerate_binary_solutions
import pyomo.contrib.alternative_solutions.tests.test_cases as tc

solvers = list(pyomo.opt.check_available_solvers("glpk", "gurobi", "appsi_gurobi"))
pytestmark = unittest.pytest.mark.parametrize("mip_solver", solvers)


@unittest.pytest.mark.default
class TestBalasUnit:

    def test_bad_solver(self, mip_solver):
        """
        Confirm that an exception is thrown with a bad solver name.
        """
        m = tc.get_triangle_ip()
        try:
            enumerate_binary_solutions(m, solver="unknown_solver")
        except pyomo.common.errors.ApplicationError as e:
            pass

    def test_ip_feasibility(self, mip_solver):
        """
        Enumerate solutions for an ip: triangle_ip.

        Check that there is just one solution when the # of binary variables is 0.
        """
        m = tc.get_triangle_ip()
        results = enumerate_binary_solutions(m, num_solutions=100, solver=mip_solver)
        assert len(results) == 1
        assert results[0].objective_value == unittest.pytest.approx(5)

    @unittest.skipIf(True, "Ignoring fragile test for solver timeout.")
    def test_no_time(self, mip_solver):
        """
        Enumerate solutions for an ip: triangle_ip.

        Check that something sensible happens when the solver times out.
        """
        m = tc.get_triangle_ip()
        with unittest.pytest.raises(Exception):
            results = enumerate_binary_solutions(
                m, num_solutions=100, solver=mip_solver, solver_options={"TimeLimit": 0}
            )

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_knapsack_all(self, mip_solver):
        """
        Enumerate solutions for a binary problem: knapsack

        """
        m = tc.get_aos_test_knapsack(
            1, weights=[3, 4, 6, 5], values=[2, 3, 1, 4], capacity=8
        )
        results = enumerate_binary_solutions(m, num_solutions=100, solver=mip_solver)
        objectives = list(
            sorted((round(result.objective[1], 2) for result in results), reverse=True)
        )
        assert_array_almost_equal(objectives, m.ranked_solution_values)
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        assert_array_almost_equal(unique_solns_by_obj, m.num_ranked_solns)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_knapsack_x0_x1(self, mip_solver):
        """
        Enumerate solutions for a binary problem: knapsack

        Check that we only see 4 solutions that enumerate alternatives of x[1] and x[1]
        """
        m = tc.get_aos_test_knapsack(
            1, weights=[3, 4, 6, 5], values=[2, 3, 1, 4], capacity=8
        )
        results = enumerate_binary_solutions(
            m, num_solutions=100, solver=mip_solver, variables=[m.x[0], m.x[1]]
        )
        objectives = list(
            sorted((round(result.objective[1], 2) for result in results), reverse=True)
        )
        assert_array_almost_equal(objectives, [6, 5, 4, 3])
        unique_solns_by_obj = [val for val in Counter(objectives).values()]
        assert_array_almost_equal(unique_solns_by_obj, [1, 1, 1, 1])

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_knapsack_optimal_3(self, mip_solver):
        """
        Enumerate solutions for a binary problem: knapsack

        """
        m = tc.get_aos_test_knapsack(
            1, weights=[3, 4, 6, 5], values=[2, 3, 1, 4], capacity=8
        )
        results = enumerate_binary_solutions(m, num_solutions=3, solver=mip_solver)
        objectives = list(
            sorted((round(result.objective[1], 2) for result in results), reverse=True)
        )
        assert_array_almost_equal(objectives, m.ranked_solution_values[:3])

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_knapsack_hamming_3(self, mip_solver):
        """
        Enumerate solutions for a binary problem: knapsack

        """
        m = tc.get_aos_test_knapsack(
            1, weights=[3, 4, 6, 5], values=[2, 3, 1, 4], capacity=8
        )
        results = enumerate_binary_solutions(
            m, num_solutions=3, solver=mip_solver, search_mode="hamming"
        )
        objectives = list(
            sorted((round(result.objective[1], 2) for result in results), reverse=True)
        )
        assert_array_almost_equal(objectives, [6, 3, 1])

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_knapsack_random_3(self, mip_solver):
        """
        Enumerate solutions for a binary problem: knapsack

        """
        m = tc.get_aos_test_knapsack(
            1, weights=[3, 4, 6, 5], values=[2, 3, 1, 4], capacity=8
        )
        results = enumerate_binary_solutions(
            m, num_solutions=3, solver=mip_solver, search_mode="random", seed=1118798374
        )
        objectives = list(
            sorted((round(result.objective[1], 2) for result in results), reverse=True)
        )
        assert_array_almost_equal(objectives, [6, 5, 4])


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""
See David Bernal PhD proposal example.
Link: https://www.researchgate.net/project/Convex-MINLP/update/5c7eb2ee3843b034242e9e4a
"""

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    Reals,
    Integers,
    Objective,
    Var,
    sqrt,
    minimize,
)
from pyomo.common.collections import ComponentMap


class ProposalModel(ConcreteModel):
    def __init__(self, *args, **kwargs):
        """Create the problem."""
        kwargs.setdefault('name', 'DavidProposalExample')
        super(ProposalModel, self).__init__(*args, **kwargs)
        m = self

        m.x = Var(domain=Reals, bounds=(0, 20), initialize=1)
        m.y = Var(domain=Integers, bounds=(0, 20), initialize=4)

        m.c1 = Constraint(expr=m.x**2 / 20.0 + m.y <= 20)
        m.c2 = Constraint(expr=(m.x - 1) ** 2 / 40.0 - m.y <= -4)
        m.c3 = Constraint(expr=m.y - 10 * sqrt(m.x + 0.1) <= 0)
        m.c4 = Constraint(expr=-m.x - m.y <= -5)

        m.objective = Objective(expr=m.x - m.y / 4.5 + 2, sense=minimize)
        m.optimal_value = 0.66555
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.x] = 1.1099999999999999
        m.optimal_solution[m.y] = 11.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import threading
import pyomo.common.unittest as unittest
from pyomo.common.multithread import *
from threading import Thread
from pyomo.opt.base.solvers import check_available_solvers


class Dummy:
    """asdfg"""

    def __init__(self):
        self.number = 1
        self.rnd = threading.get_ident()

    def __str__(self):
        return 'asd'


class TestMultithreading(unittest.TestCase):
    def test_wrapper_docs(self):
        sut = MultiThreadWrapper(Dummy)
        self.assertEqual(sut.__class__.__doc__, Dummy.__doc__)

    def test_wrapper_field(self):
        sut = MultiThreadWrapper(Dummy)
        self.assertEqual(sut.number, 1)

    def test_independent_writes(self):
        sut = MultiThreadWrapper(Dummy)
        sut.number = 2

        def thread_func():
            self.assertEqual(sut.number, 1)

        t = Thread(target=thread_func)
        t.start()
        t.join()

    def test_independent_writes2(self):
        sut = MultiThreadWrapper(Dummy)

        def thread_func():
            sut.number = 2

        t = Thread(target=thread_func)
        t.start()
        t.join()
        self.assertEqual(sut.number, 1)

    def test_independent_del(self):
        sut = MultiThreadWrapper(Dummy)
        del sut.number

        def thread_func():
            self.assertEqual(sut.number, 1)

        t = Thread(target=thread_func)
        t.start()
        t.join()

    def test_independent_del2(self):
        sut = MultiThreadWrapper(Dummy)

        def thread_func():
            del sut.number

        t = Thread(target=thread_func)
        t.start()
        t.join()
        self.assertEqual(sut.number, 1)

    def test_special_methods(self):
        sut = MultiThreadWrapper(Dummy)
        self.assertTrue(set(Dummy().__dir__()).issubset(set(sut.__dir__())))
        self.assertEqual(str(sut), str(Dummy()))

    def test_main(self):
        sut = MultiThreadWrapperWithMain(Dummy)
        self.assertIs(sut.main_thread.rnd, sut.rnd)
        sut.number = 5

        def thread_func():
            self.assertEqual(sut.number, 1)
            self.assertIsNot(sut.main_thread.rnd, sut.rnd)
            del sut.number

        t = Thread(target=thread_func)
        t.start()
        t.join()
        self.assertEqual(sut.number, 5)

    @unittest.skipIf(
        len(check_available_solvers('glpk')) < 1, "glpk solver not available"
    )
    def test_solve(self):
        # Based on the minimal example in https://github.com/Pyomo/pyomo/issues/2475
        import pyomo.environ as pyo
        from pyomo.opt import SolverFactory
        from multiprocessing.dummy import Pool as ThreadPool

        model = pyo.ConcreteModel()
        model.nVars = pyo.Param(initialize=4)
        model.N = pyo.RangeSet(model.nVars)
        model.x = pyo.Var(model.N, within=pyo.Binary)
        model.obj = pyo.Objective(expr=pyo.summation(model.x))
        model.cuts = pyo.ConstraintList()

        def test(model):
            opt = SolverFactory('glpk')
            opt.solve(model)

            # Iterate, adding a cut to exclude the previously found solution
            for _ in range(5):
                expr = 0
                for j in model.x:
                    if pyo.value(model.x[j]) < 0.5:
                        expr += model.x[j]
                    else:
                        expr += 1 - model.x[j]
                model.cuts.add(expr >= 1)
                results = opt.solve(model)
            return results, [v for v in model.x]

        tp = ThreadPool(4)
        results = tp.map(test, [model.clone() for i in range(4)])
        tp.close()
        for result, _ in results:
            self.assertEqual(
                result.solver.termination_condition, pyo.TerminationCondition.optimal
            )
        results = list(results)
        for _, values in results[1:]:
            self.assertEqual(values, results[0][1])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
import os
import pyomo.common.unittest as unittest
from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy_available,
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")
from pyomo.contrib.pynumero.asl import AmplInterface

if not AmplInterface.available():
    raise unittest.SkipTest("Pynumero needs the ASL extension to run NLP tests")
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.common.gsl import find_GSL
from pyomo.environ import ConcreteModel, ExternalFunction, Var, Objective


class TestAMPLExternalFunction(unittest.TestCase):
    def assertListsAlmostEqual(self, first, second, places=7, msg=None):
        self.assertEqual(len(first), len(second))
        msg = "lists %s and %s differ at item " % (first, second)
        for i, a in enumerate(first):
            self.assertAlmostEqual(a, second[i], places, msg + str(i))

    def test_solve_gsl_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest("Could not find the amplgsl.dll library")
        model = ConcreteModel()
        model.z_func = ExternalFunction(library=DLL, function="gsl_sf_gamma")
        model.x = Var(initialize=3, bounds=(1e-5, None))
        model.o = Objective(expr=model.z_func(model.x))
        nlp = PyomoNLP(model)
        self.assertAlmostEqual(nlp.evaluate_objective(), 2, 7)
        assert "AMPLFUNC" not in os.environ


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.tee import capture_output

import gc
from io import StringIO
from itertools import zip_longest
import logging
import sys
import time

from pyomo.common.log import LoggingIntercept
from pyomo.common.timing import (
    ConstructionTimer,
    TransformationTimer,
    report_timing,
    TicTocTimer,
    HierarchicalTimer,
)
from pyomo.environ import (
    AbstractModel,
    ConcreteModel,
    RangeSet,
    Var,
    Any,
    TransformationFactory,
)
from pyomo.core.base.var import VarData


class _pseudo_component(Var):
    def getname(*args, **kwds):
        raise RuntimeError("fail")


class TestTiming(unittest.TestCase):
    def setUp(self):
        self.reenable_gc = gc.isenabled()
        gc.disable()
        # Set a long switch interval to discourage context switches
        # during these tests
        self.switchinterval = sys.getswitchinterval()
        sys.setswitchinterval(10)

    def tearDown(self):
        sys.setswitchinterval(self.switchinterval)
        if self.reenable_gc:
            gc.enable()
            gc.collect()

    def test_raw_construction_timer(self):
        a = ConstructionTimer(None)
        self.assertRegex(
            str(a),
            r"ConstructionTimer object for NoneType \(unknown\); "
            r"[0-9\.]+ elapsed seconds",
        )
        v = Var()
        v.construct()
        a = ConstructionTimer(VarData(v))
        self.assertRegex(
            str(a),
            r"ConstructionTimer object for Var ScalarVar\[NOTSET\]; "
            r"[0-9\.]+ elapsed seconds",
        )

    def test_raw_transformation_timer(self):
        a = TransformationTimer(None)
        self.assertRegex(
            str(a), r"TransformationTimer object for NoneType; [0-9\.]+ elapsed seconds"
        )

        v = _pseudo_component()
        a = ConstructionTimer(v)
        self.assertIn("ConstructionTimer object for Var (unknown); ", str(a))

    def test_raw_transformation_timer(self):
        a = TransformationTimer(None, 'fwd')
        self.assertIn("TransformationTimer object for NoneType (fwd); ", str(a))

        a = TransformationTimer(None)
        self.assertIn("TransformationTimer object for NoneType; ", str(a))

    def test_report_timing(self):
        ref = r"""
           (0(\.\d+)?) seconds to construct Block ConcreteModel; 1 index total
           (0(\.\d+)?) seconds to construct RangeSet FiniteScalarRangeSet; 1 index total
           (0(\.\d+)?) seconds to construct Var x; 2 indices total
           (0(\.\d+)?) seconds to construct Var y; 0 indices total
           (0(\.\d+)?) seconds to construct Suffix Suffix
           (0(\.\d+)?) seconds to apply Transformation RelaxIntegerVars \(in-place\)
           """.strip()

        xfrm = TransformationFactory('core.relax_integer_vars')

        try:
            with capture_output() as out:
                report_timing()
                m = ConcreteModel()
                m.r = RangeSet(2)
                m.x = Var(m.r)
                m.y = Var(Any, dense=False)
                xfrm.apply_to(m)
            result = out.getvalue().strip()
            for l, r in zip(result.splitlines(), ref.splitlines()):
                self.assertRegex(str(l.strip()), str(r.strip()))
        finally:
            report_timing(False)

        os = StringIO()
        try:
            report_timing(os)
            m = ConcreteModel()
            m.r = RangeSet(2)
            m.x = Var(m.r)
            m.y = Var(Any, dense=False)
            xfrm.apply_to(m)
            result = os.getvalue().strip()
            for l, r in zip(result.splitlines(), ref.splitlines()):
                self.assertRegex(str(l.strip()), str(r.strip()))
        finally:
            report_timing(False)
        buf = StringIO()
        with LoggingIntercept(buf, 'pyomo'):
            m = ConcreteModel()
            m.r = RangeSet(2)
            m.x = Var(m.r)
            m.y = Var(Any, dense=False)
            xfrm.apply_to(m)
            result = os.getvalue().strip()
            for l, r in zip(result.splitlines(), ref.splitlines()):
                self.assertRegex(str(l.strip()), str(r.strip()))
            self.assertEqual(buf.getvalue().strip(), "")

    def test_report_timing_context_manager(self):
        ref = r"""
           (0(\.\d+)?) seconds to construct Var x; 2 indices total
           (0(\.\d+)?) seconds to construct Var y; 0 indices total
           (0(\.\d+)?) seconds to construct Suffix Suffix
           (0(\.\d+)?) seconds to apply Transformation RelaxIntegerVars \(in-place\)
           """.strip()

        xfrm = TransformationFactory('core.relax_integer_vars')

        model = AbstractModel()
        model.r = RangeSet(2)
        model.x = Var(model.r)
        model.y = Var(Any, dense=False)

        OS = StringIO()

        with report_timing(False):
            with report_timing(OS):
                with report_timing(False):
                    # Active reporting is False: nothing should be emitted
                    with capture_output() as OUT:
                        m = model.create_instance()
                        xfrm.apply_to(m)
                    self.assertEqual(OUT.getvalue(), "")
                    self.assertEqual(OS.getvalue(), "")
                # Active reporting: we should log the timing
                with capture_output() as OUT:
                    m = model.create_instance()
                    xfrm.apply_to(m)
                self.assertEqual(OUT.getvalue(), "")
                result = OS.getvalue().strip()
                for l, r in zip_longest(result.splitlines(), ref.splitlines()):
                    self.assertRegex(str(l.strip()), str(r.strip()))
            # Active reporting is False: the previous log should not have changed
            with capture_output() as OUT:
                m = model.create_instance()
                xfrm.apply_to(m)
            self.assertEqual(OUT.getvalue(), "")
            self.assertEqual(result, OS.getvalue().strip())

    def test_TicTocTimer_tictoc(self):
        SLEEP = 0.1
        RES = 0.01  # resolution (seconds): 1/10 the sleep

        # Note: pypy on GHA occasionally has timing
        # differences of >0.04s
        if 'pypy_version_info' in dir(sys):
            RES *= 2.5
        # Note: previously, OSX on GHA also had significantly nosier tests
        # if sys.platform == 'darwin':
        #     RES *= 2

        # Note: the above RES heuristics were determined before the
        # current handling of "now" within the tic/toc timer.  They are
        # probably overly conservative now, but tightening them doesn't
        # really improve the quality of the tests.

        abs_time = time.perf_counter()
        timer = TicTocTimer()

        time.sleep(SLEEP)

        with capture_output() as out:
            start_time = time.perf_counter()
            timer.tic(None)
        self.assertEqual(out.getvalue(), '')

        with capture_output() as out:
            start_time = time.perf_counter()
            timer.tic()
        self.assertRegex(
            out.getvalue(), r'\[    [.0-9]+\] Resetting the tic/toc delta timer'
        )

        time.sleep(SLEEP)

        with capture_output() as out:
            ref = time.perf_counter()
            delta = timer.toc()
        self.assertAlmostEqual(ref - start_time, delta, delta=RES)
        self.assertRegex(
            out.getvalue(), r'\[\+   [.0-9]+\] .* in test_TicTocTimer_tictoc'
        )
        with capture_output() as out:
            # entering / leaving the context manager can take non-trivial
            # time on some platforms (up to 0.03 on Windows / Python 3.10)
            self.assertAlmostEqual(
                time.perf_counter() - ref, timer.toc(None), delta=RES
            )
        self.assertEqual(out.getvalue(), '')

        with capture_output() as out:
            ref = time.perf_counter()
            total = timer.toc(delta=False)
        self.assertAlmostEqual(ref - start_time, total, delta=RES)
        self.assertRegex(
            out.getvalue(), r'\[    [.0-9]+\] .* in test_TicTocTimer_tictoc'
        )

        ref *= -1
        time.sleep(SLEEP)

        ref += time.perf_counter()
        timer.stop()
        cumul_stop1 = timer.toc(None)
        self.assertAlmostEqual(ref, cumul_stop1, delta=RES)
        with self.assertRaisesRegex(
            RuntimeError, 'Stopping a TicTocTimer that was already stopped'
        ):
            timer.stop()
        time.sleep(SLEEP)
        cumul_stop2 = timer.toc(None)
        self.assertEqual(cumul_stop1, cumul_stop2)

        ref -= time.perf_counter()
        timer.start()
        time.sleep(SLEEP)

        with capture_output() as out:
            ref += time.perf_counter()
            timer.stop()
            delta = timer.toc()
        self.assertAlmostEqual(ref, delta, delta=RES)
        self.assertRegex(
            out.getvalue(), r'\[    [.0-9]+\|   1\] .* in test_TicTocTimer_tictoc'
        )
        with capture_output() as out:
            # Note that delta is ignored if the timer is a cumulative timer
            total = timer.toc(delta=False)
        self.assertAlmostEqual(delta, total, delta=RES)
        self.assertRegex(
            out.getvalue(), r'\[    [.0-9]+\|   1\] .* in test_TicTocTimer_tictoc'
        )

    def test_TicTocTimer_context_manager(self):
        SLEEP = 0.1
        RES = 0.01  # resolution (seconds): 1/10 the sleep

        # Note: the above RES heuristic was determined before the
        # current handling of "now" within the tic/toc timer.  It is
        # probably overly conservative now, but tightening them doesn't
        # really improve the quality of the tests.

        abs_time = time.perf_counter()
        with TicTocTimer() as timer:
            time.sleep(SLEEP)
        exclude = -time.perf_counter()
        time.sleep(SLEEP)
        exclude += time.perf_counter()
        with timer:
            time.sleep(SLEEP)
        abs_time = time.perf_counter() - abs_time
        self.assertGreater(abs_time, SLEEP * 3 - RES / 10)
        self.assertAlmostEqual(timer.toc(None), abs_time - exclude, delta=RES)

    def test_TicTocTimer_logger(self):
        # test specifying a logger to the timer constructor disables all
        # output to stdout
        timer = TicTocTimer(logger=logging.getLogger(__name__))
        with capture_output() as OUT:
            with LoggingIntercept(level=logging.INFO) as LOG:
                timer.toc("msg1")
            self.assertRegex(LOG.getvalue(), r"^\[[^\]]+\] msg1$")
            with LoggingIntercept(level=logging.INFO) as LOG:
                timer.toc("msg2", level=logging.DEBUG)
            self.assertEqual(LOG.getvalue(), "")
        self.assertEqual(OUT.getvalue(), "")

        with capture_output() as OUT:
            with LoggingIntercept(level=logging.DEBUG) as LOG:
                timer.toc("msg1")
                timer.toc("msg2", level=logging.DEBUG)
        self.assertEqual(OUT.getvalue(), "")
        self.assertRegex(LOG.getvalue(), r"^\[[^\]]+\] msg1\n\[[^\]]+\] msg2$")

        # test specifying a logger to toc() for a general timer disables
        # output to stdout
        timer = TicTocTimer()
        with capture_output() as OUT:
            with LoggingIntercept(level=logging.INFO) as LOG:
                timer.toc("msg1", logger=logging.getLogger(__name__))
        self.assertRegex(LOG.getvalue(), r"^\[[^\]]+\] msg1$")
        self.assertEqual(OUT.getvalue(), "")

    def test_TicTocTimer_deprecated(self):
        timer = TicTocTimer()
        with LoggingIntercept() as LOG, capture_output() as out:
            timer.tic("msg", None, None)
        self.assertEqual(out.getvalue(), "")
        self.assertRegex(
            LOG.getvalue().replace('\n', ' ').strip(),
            r"DEPRECATED: tic\(\): 'ostream' and 'logger' should be specified "
            r"as keyword arguments( +\([^\)]+\)){2}",
        )

        with LoggingIntercept() as LOG, capture_output() as out:
            timer.toc("msg", True, None, None)
        self.assertEqual(out.getvalue(), "")
        self.assertRegex(
            LOG.getvalue().replace('\n', ' ').strip(),
            r"DEPRECATED: toc\(\): 'delta', 'ostream', and 'logger' should be "
            r"specified as keyword arguments( +\([^\)]+\)){2}",
        )

        timer = TicTocTimer()
        with LoggingIntercept() as LOG, capture_output() as out:
            timer.tic("msg %s, %s", None, None)
        self.assertIn('msg None, None', out.getvalue())
        self.assertEqual(LOG.getvalue(), "")

        with LoggingIntercept() as LOG, capture_output() as out:
            timer.toc("msg %s, %s, %s", True, None, None)
        self.assertIn('msg True, None, None', out.getvalue())
        self.assertEqual(LOG.getvalue(), "")

    def test_HierarchicalTimer(self):
        RES = 0.01  # resolution (seconds)

        timer = HierarchicalTimer()
        start_time = time.perf_counter()
        timer.start('all')
        time.sleep(0.02)
        for i in range(10):
            timer.start('a')
            time.sleep(0.01)
            for j in range(5):
                timer.start('aa')
                time.sleep(0.001)
                timer.stop('aa')
            timer.start('ab')
            timer.stop('ab')
            timer.stop('a')
        end_time = time.perf_counter()
        timer.stop('all')
        ref = """Identifier        ncalls   cumtime   percall      %
---------------------------------------------------
all                    1     [0-9.]+ +[0-9.]+ +100.0
     ----------------------------------------------
     a                10     [0-9.]+ +[0-9.]+ +[0-9.]+
          -----------------------------------------
          aa          50     [0-9.]+ +[0-9.]+ +[0-9.]+
          ab          10     [0-9.]+ +[0-9.]+ +[0-9.]+
          other      n/a     [0-9.]+ +n/a +[0-9.]+
          =========================================
     other           n/a     [0-9.]+ +n/a +[0-9.]+
     ==============================================
===================================================
""".splitlines()
        for l, r in zip(str(timer).splitlines(), ref):
            self.assertRegex(l, r)

        self.assertEqual(1, timer.get_num_calls('all'))
        self.assertAlmostEqual(
            end_time - start_time, timer.get_total_time('all'), delta=RES
        )
        self.assertEqual(100.0, timer.get_relative_percent_time('all'))
        self.assertTrue(100.0 > timer.get_relative_percent_time('all.a'))
        self.assertTrue(50.0 < timer.get_relative_percent_time('all.a'))

    def test_HierarchicalTimer_longNames(self):
        RES = 0.01  # resolution (seconds)

        timer = HierarchicalTimer()
        start_time = time.perf_counter()
        timer.start('all' * 25)
        time.sleep(0.02)
        for i in range(10):
            timer.start('a' * 75)
            time.sleep(0.01)
            for j in range(5):
                timer.start('aa' * 20)
                time.sleep(0.001)
                timer.stop('aa' * 20)
            timer.start('ab' * 20)
            timer.stop('ab' * 20)
            timer.stop('a' * 75)
        end_time = time.perf_counter()
        timer.stop('all' * 25)
        ref = (
            """Identifier%s   ncalls   cumtime   percall      %%
%s------------------------------------
%s%s        1     [0-9.]+ +[0-9.]+ +100.0
    %s------------------------------------
    %s%s       10     [0-9.]+ +[0-9.]+ +[0-9.]+
        %s------------------------------------
        %s%s       50     [0-9.]+ +[0-9.]+ +[0-9.]+
        %s%s       10     [0-9.]+ +[0-9.]+ +[0-9.]+
        other%s      n/a     [0-9.]+ +n/a +[0-9.]+
        %s====================================
    other%s      n/a     [0-9.]+ +n/a +[0-9.]+
    %s====================================
%s====================================
"""
            % (
                ' ' * 69,
                '-' * 79,
                'all' * 25,
                ' ' * 4,
                '-' * 75,
                'a' * 75,
                '',
                '-' * 71,
                'aa' * 20,
                ' ' * 31,
                'ab' * 20,
                ' ' * 31,
                ' ' * 66,
                '=' * 71,
                ' ' * 70,
                '=' * 75,
                '=' * 79,
            )
        ).splitlines()
        for l, r in zip(str(timer).splitlines(), ref):
            self.assertRegex(l, r)

    def test_clear_except_base_timer(self):
        timer = HierarchicalTimer()
        timer.start("a")
        timer.start("b")
        timer.stop("b")
        timer.stop("a")
        timer.start("c")
        timer.stop("c")
        timer.start("d")
        timer.stop("d")
        timer.clear_except("b", "c")
        key_set = set(timer.timers.keys())
        self.assertEqual(key_set, {"c"})

    def test_clear_except_subtimer(self):
        # Testing this method on "sub-timers" exercises different code
        # as while the base timer is a HierarchicalTimer, the sub-timers
        # are _HierarchicalHelpers
        timer = HierarchicalTimer()
        timer.start("root")
        timer.start("a")
        timer.start("b")
        timer.stop("b")
        timer.stop("a")
        timer.start("c")
        timer.stop("c")
        timer.start("d")
        timer.stop("d")
        timer.stop("root")
        root = timer.timers["root"]
        root.clear_except("b", "c")
        key_set = set(root.timers.keys())
        self.assertEqual(key_set, {"c"})


class TestFlattenHierarchicalTimer(unittest.TestCase):
    #
    # The following methods create some hierarchical timers, then
    # hand-code the total time of each timer in the data structure.
    # This is so we can assert that total_time fields of flattened
    # timers are computed correctly without relying on the actual
    # time spent.
    #
    def make_singleton_timer(self):
        timer = HierarchicalTimer()
        timer.start("root")
        timer.stop("root")
        timer.timers["root"].total_time = 5.0
        return timer

    def make_flat_timer(self):
        timer = HierarchicalTimer()
        timer.start("root")
        timer.start("a")
        timer.stop("a")
        timer.start("b")
        timer.stop("b")
        timer.stop("root")
        timer.timers["root"].total_time = 5.0
        timer.timers["root"].timers["a"].total_time = 1.0
        timer.timers["root"].timers["b"].total_time = 2.5
        return timer

    def make_timer_depth_2_one_child(self):
        timer = HierarchicalTimer()
        timer.start("root")
        timer.start("a")
        timer.start("b")
        timer.stop("b")
        timer.start("c")
        timer.stop("c")
        timer.stop("a")
        timer.stop("root")
        timer.timers["root"].total_time = 5.0
        timer.timers["root"].timers["a"].total_time = 4.0
        timer.timers["root"].timers["a"].timers["b"].total_time = 1.1
        timer.timers["root"].timers["a"].timers["c"].total_time = 2.2
        return timer

    def make_timer_depth_2_with_name_collision(self):
        timer = HierarchicalTimer()
        timer.start("root")
        timer.start("a")
        timer.start("b")
        timer.stop("b")
        timer.start("c")
        timer.stop("c")
        timer.stop("a")
        timer.start("b")
        timer.stop("b")
        timer.stop("root")
        timer.timers["root"].total_time = 5.0
        timer.timers["root"].timers["a"].total_time = 4.0
        timer.timers["root"].timers["a"].timers["b"].total_time = 1.1
        timer.timers["root"].timers["a"].timers["c"].total_time = 2.2
        timer.timers["root"].timers["b"].total_time = 0.11
        return timer

    def make_timer_depth_2_two_children(self):
        timer = HierarchicalTimer()
        timer.start("root")
        timer.start("a")
        timer.start("b")
        timer.stop("b")
        timer.start("c")
        timer.stop("c")
        timer.stop("a")
        timer.start("b")
        timer.start("c")
        timer.stop("c")
        timer.start("d")
        timer.stop("d")
        timer.stop("b")
        timer.stop("root")
        timer.timers["root"].total_time = 5.0
        timer.timers["root"].timers["a"].total_time = 4.0
        timer.timers["root"].timers["a"].timers["b"].total_time = 1.1
        timer.timers["root"].timers["a"].timers["c"].total_time = 2.2
        timer.timers["root"].timers["b"].total_time = 0.88
        timer.timers["root"].timers["b"].timers["c"].total_time = 0.07
        timer.timers["root"].timers["b"].timers["d"].total_time = 0.05
        return timer

    def make_timer_depth_4(self):
        timer = HierarchicalTimer()
        timer.start("root")
        timer.start("a")
        timer.start("b")
        timer.stop("b")
        timer.start("c")
        timer.start("d")
        timer.start("e")
        timer.stop("e")
        timer.stop("d")
        timer.stop("c")
        timer.stop("a")
        timer.start("b")
        timer.start("c")
        timer.start("e")
        timer.stop("e")
        timer.stop("c")
        timer.start("d")
        timer.stop("d")
        timer.stop("b")
        timer.stop("root")
        timer.timers["root"].total_time = 5.0
        timer.timers["root"].timers["a"].total_time = 4.0
        timer.timers["root"].timers["a"].timers["b"].total_time = 1.1
        timer.timers["root"].timers["a"].timers["c"].total_time = 2.2
        timer.timers["root"].timers["a"].timers["c"].timers["d"].total_time = 0.9
        timer.timers["root"].timers["a"].timers["c"].timers["d"].timers[
            "e"
        ].total_time = 0.6
        timer.timers["root"].timers["b"].total_time = 0.88
        timer.timers["root"].timers["b"].timers["c"].total_time = 0.07
        timer.timers["root"].timers["b"].timers["c"].timers["e"].total_time = 0.04
        timer.timers["root"].timers["b"].timers["d"].total_time = 0.05
        return timer

    def make_timer_depth_4_same_name(self):
        timer = HierarchicalTimer()
        timer.start("root")
        timer.start("a")
        timer.start("a")
        timer.start("a")
        timer.start("a")
        timer.stop("a")
        timer.stop("a")
        timer.stop("a")
        timer.stop("a")
        timer.stop("root")
        timer.timers["root"].total_time = 5.0
        timer.timers["root"].timers["a"].total_time = 1.0
        timer.timers["root"].timers["a"].timers["a"].total_time = 0.1
        timer.timers["root"].timers["a"].timers["a"].timers["a"].total_time = 0.01
        timer.timers["root"].timers["a"].timers["a"].timers["a"].timers[
            "a"
        ].total_time = 0.001
        return timer

    def test_singleton(self):
        timer = self.make_singleton_timer()
        root = timer.timers["root"]
        root.flatten()
        self.assertAlmostEqual(root.total_time, 5.0)

    def test_already_flat(self):
        timer = self.make_flat_timer()
        root = timer.timers["root"]
        root.flatten()
        self.assertAlmostEqual(root.total_time, 5.0)
        self.assertAlmostEqual(root.timers["a"].total_time, 1.0)
        self.assertAlmostEqual(root.timers["b"].total_time, 2.5)

    def test_depth_2_one_child(self):
        timer = self.make_timer_depth_2_one_child()
        root = timer.timers["root"]
        root.flatten()
        self.assertAlmostEqual(root.total_time, 5.0)
        self.assertAlmostEqual(root.timers["a"].total_time, 0.7)
        self.assertAlmostEqual(root.timers["b"].total_time, 1.1)
        self.assertAlmostEqual(root.timers["c"].total_time, 2.2)

    def test_timer_depth_2_with_name_collision(self):
        timer = self.make_timer_depth_2_with_name_collision()
        root = timer.timers["root"]
        root.flatten()
        self.assertAlmostEqual(root.total_time, 5.0)
        self.assertAlmostEqual(root.timers["a"].total_time, 0.700)
        self.assertAlmostEqual(root.timers["b"].total_time, 1.210)
        self.assertAlmostEqual(root.timers["c"].total_time, 2.200)

    def test_timer_depth_2_two_children(self):
        timer = self.make_timer_depth_2_two_children()
        root = timer.timers["root"]
        root.flatten()
        self.assertAlmostEqual(root.total_time, 5.0)
        self.assertAlmostEqual(root.timers["a"].total_time, 0.700)
        self.assertAlmostEqual(root.timers["b"].total_time, 1.860)
        self.assertAlmostEqual(root.timers["c"].total_time, 2.270)
        self.assertAlmostEqual(root.timers["d"].total_time, 0.050)

    def test_timer_depth_4(self):
        timer = self.make_timer_depth_4()
        root = timer.timers["root"]
        root.flatten()
        self.assertAlmostEqual(root.total_time, 5.0)
        self.assertAlmostEqual(root.timers["a"].total_time, 0.700)
        self.assertAlmostEqual(root.timers["b"].total_time, 1.860)
        self.assertAlmostEqual(root.timers["c"].total_time, 1.330)
        self.assertAlmostEqual(root.timers["d"].total_time, 0.350)
        self.assertAlmostEqual(root.timers["e"].total_time, 0.640)

    def test_timer_depth_4_same_name(self):
        timer = self.make_timer_depth_4_same_name()
        root = timer.timers["root"]
        root.flatten()
        self.assertAlmostEqual(root.total_time, 5.0)
        self.assertAlmostEqual(root.timers["a"].total_time, 1.0)

    def test_base_timer_depth_3(self):
        # This is depth 2 wrt the root, depth 3 wrt the
        # "base timer"
        timer = self.make_timer_depth_2_two_children()
        timer.flatten()
        self.assertAlmostEqual(timer.timers["root"].total_time, 0.120)
        self.assertAlmostEqual(timer.timers["a"].total_time, 0.700)
        self.assertAlmostEqual(timer.timers["b"].total_time, 1.860)
        self.assertAlmostEqual(timer.timers["c"].total_time, 2.270)
        self.assertAlmostEqual(timer.timers["d"].total_time, 0.050)

    def test_timer_still_active(self):
        timer = HierarchicalTimer()
        timer.start("a")
        timer.stop("a")
        timer.start("b")
        msg = "Cannot flatten.*while any timers are active"
        with self.assertRaisesRegex(RuntimeError, msg):
            timer.flatten()
        timer.stop("b")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

from pyomo.common.dependencies import numpy as np, scipy_available, numpy_available
import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    inequality,
    Objective,
    maximize,
    Var,
)
from pyomo.core.expr import (
    MonomialTermExpression,
    NegationExpression,
    ProductExpression,
)
from pyomo.core.expr.compare import (
    assertExpressionsEqual,
    assertExpressionsStructurallyEqual,
)

from pyomo.repn.plugins.parameterized_standard_form import (
    ParameterizedLinearStandardFormCompiler,
    _CSRMatrix,
    _CSCMatrix,
)


@unittest.skipUnless(
    numpy_available & scipy_available,
    "CSC and CSR representations require scipy and numpy",
)
class TestSparseMatrixRepresentations(unittest.TestCase):
    def test_csr_to_csc_only_data(self):
        A = _CSRMatrix(([5, 8, 3, 6], [0, 1, 2, 1], [0, 1, 2, 3, 4]), [4, 4])
        thing = A.tocsc()

        self.assertTrue(np.all(thing.data == np.array([5, 8, 6, 3])))
        self.assertTrue(np.all(thing.indices == np.array([0, 1, 3, 2])))
        self.assertTrue(np.all(thing.indptr == np.array([0, 1, 3, 4, 4])))

        should_be_A = thing.tocsr()
        self.assertTrue(np.all(should_be_A.data == A.data))
        self.assertTrue(np.all(should_be_A.indices == A.indices))
        self.assertTrue(np.all(should_be_A.indptr == A.indptr))

    def test_csr_to_csc_pyomo_exprs(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        A = _CSRMatrix(
            ([5, 8 * m.x, 3 * m.x * m.y**2, 6], [0, 1, 2, 1], [0, 1, 2, 3, 4]), [4, 4]
        )
        thing = A.tocsc()

        self.assertEqual(thing.data[0], 5)
        assertExpressionsEqual(self, thing.data[1], 8 * m.x)
        self.assertEqual(thing.data[2], 6)
        assertExpressionsEqual(self, thing.data[3], 3 * m.x * m.y**2)
        self.assertEqual(thing.data.shape, (4,))

        self.assertTrue(np.all(thing.indices == np.array([0, 1, 3, 2])))
        self.assertTrue(np.all(thing.indptr == np.array([0, 1, 3, 4, 4])))

        should_be_A = thing.tocsr()
        self.assertEqual(should_be_A.data[0], 5)
        assertExpressionsEqual(self, should_be_A.data[1], 8 * m.x)
        assertExpressionsEqual(self, should_be_A.data[2], 3 * m.x * m.y**2)
        self.assertEqual(should_be_A.data[3], 6)
        self.assertEqual(should_be_A.data.shape, (4,))
        self.assertTrue(np.all(should_be_A.indices == np.array([0, 1, 2, 1])))
        self.assertTrue(np.all(should_be_A.indptr == np.array([0, 1, 2, 3, 4])))

    def test_csr_to_csc_empty_matrix(self):
        A = _CSRMatrix(([], [], [0]), [0, 4])
        thing = A.tocsc()

        self.assertEqual(thing.data.size, 0)
        self.assertEqual(thing.indices.size, 0)
        self.assertEqual(thing.shape, (0, 4))
        self.assertTrue(np.all(thing.indptr == np.zeros(5)))

        should_be_A = thing.tocsr()
        self.assertEqual(should_be_A.data.size, 0)
        self.assertEqual(should_be_A.indices.size, 0)
        self.assertEqual(should_be_A.shape, (0, 4))
        self.assertTrue(np.all(should_be_A.indptr == np.zeros(5)))

    def test_todense(self):
        A = _CSRMatrix(([5, 8, 3, 6], [0, 1, 2, 1], [0, 1, 2, 3, 4]), [4, 4])
        dense = np.array([[5, 0, 0, 0], [0, 8, 0, 0], [0, 0, 3, 0], [0, 6, 0, 0]])

        self.assertTrue(np.all(A.todense() == dense))
        self.assertTrue(np.all(A.tocsc().todense() == dense))

        A = _CSRMatrix(
            ([5, 6, 7, 2, 1, 1.5], [0, 1, 1, 2, 3, 1], [0, 2, 4, 5, 6]), [4, 4]
        )
        dense = np.array([[5, 6, 0, 0], [0, 7, 2, 0], [0, 0, 0, 1], [0, 1.5, 0, 0]])
        self.assertTrue(np.all(A.todense() == dense))
        self.assertTrue(np.all(A.tocsc().todense() == dense))

    def test_sum_duplicates(self):
        A = _CSCMatrix(([4, 5], [1, 1], [0, 0, 2, 2]), [3, 3])
        self.assertTrue(np.all(A.data == [4, 5]))
        self.assertTrue(np.all(A.indptr == [0, 0, 2, 2]))
        self.assertTrue(np.all(A.indices == [1, 1]))

        A.sum_duplicates()

        self.assertTrue(np.all(A.data == [9]))
        self.assertTrue(np.all(A.indptr == [0, 0, 1, 1]))
        self.assertTrue(np.all(A.indices == [1]))

        dense = np.array([[0, 0, 0], [0, 9, 0], [0, 0, 0]])
        self.assertTrue(np.all(A.todense() == dense))

    def test_invalid_sparse_matrix_input(self):
        with self.assertRaisesRegex(
            ValueError,
            r"Shape specifies the number of rows as 3 but the index "
            r"pointer has length 2. The index pointer must have length "
            r"nrows \+ 1: Check the 'shape' and 'matrix_data' arguments.",
        ):
            A = _CSRMatrix(([4, 5], [1, 1], [1, 1]), shape=(3, 3))

        with self.assertRaisesRegex(
            ValueError,
            r"Shape specifies the number of columns as 3 but the index "
            r"pointer has length 2. The index pointer must have length "
            r"ncols \+ 1: Check the 'shape' and 'matrix_data' arguments.",
        ):
            A = _CSCMatrix(([4, 5], [1, 1], [1, 1]), shape=(3, 3))


def assertExpressionArraysEqual(self, A, B):
    self.assertEqual(A.shape, B.shape)
    for i in range(A.shape[0]):
        for j in range(A.shape[1]):
            assertExpressionsStructurallyEqual(self, A[i, j], B[i, j])


def assertExpressionListsEqual(self, A, B):
    self.assertEqual(len(A), len(B))
    for i, a in enumerate(A):
        assertExpressionsEqual(self, a, B[i])


@unittest.skipUnless(
    numpy_available & scipy_available,
    "Parameterized standard form requires scipy and numpy",
)
class TestParameterizedStandardFormCompiler(unittest.TestCase):
    def test_linear_model(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])
        m.c = Constraint(expr=m.x + 2 * m.y[1] >= 3)
        m.d = Constraint(expr=m.y[1] + 4 * m.y[3] <= 5)

        repn = ParameterizedLinearStandardFormCompiler().write(m)

        self.assertTrue(np.all(repn.c == np.array([0, 0, 0])))
        self.assertTrue(np.all(repn.A == np.array([[-1, -2, 0], [0, 1, 4]])))
        self.assertTrue(np.all(repn.rhs == np.array([-3, 5])))
        self.assertEqual(repn.rows, [(m.c, -1), (m.d, 1)])
        self.assertEqual(repn.columns, [m.x, m.y[1], m.y[3]])

    def test_parameterized_linear_model(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])
        m.data = Var([1, 2])
        m.more_data = Var()
        m.c = Constraint(expr=m.x + 2 * m.data[1] * m.data[2] * m.y[1] >= 3)
        m.d = Constraint(expr=m.y[1] + 4 * m.y[3] <= 5 * m.more_data)

        repn = ParameterizedLinearStandardFormCompiler().write(
            m, wrt=[m.data, m.more_data]
        )

        self.assertTrue(np.all(repn.c == np.array([0, 0, 0])))
        assertExpressionArraysEqual(
            self,
            repn.A.todense(),
            np.array(
                [
                    [
                        -1,
                        NegationExpression(
                            (
                                ProductExpression(
                                    [MonomialTermExpression([2, m.data[1]]), m.data[2]]
                                ),
                            )
                        ),
                        0,
                    ],
                    [0, 1, 4],
                ]
            ),
        )
        assertExpressionListsEqual(self, repn.rhs, [-3, 5 * m.more_data])
        self.assertEqual(repn.rows, [(m.c, -1), (m.d, 1)])
        self.assertEqual(repn.columns, [m.x, m.y[1], m.y[3]])

    def test_parameterized_almost_dense_linear_model(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])
        m.data = Var([1, 2])
        m.more_data = Var()
        m.c = Constraint(
            expr=m.x + 2 * m.y[1] + 4 * m.y[3] + m.more_data >= 10 * m.data[1] ** 2
        )
        m.d = Constraint(expr=5 * m.x + 6 * m.y[1] + 8 * m.data[2] * m.y[3] <= 20)

        repn = ParameterizedLinearStandardFormCompiler().write(
            m, wrt=[m.data, m.more_data]
        )

        self.assertTrue(np.all(repn.c == np.array([0, 0, 0])))
        # m.c gets interpreted as a <= Constraint, and you can't really blame
        # pyomo for that because it's not parameterized yet. So that's why this
        # differs from the test in test_standard_form.py
        assertExpressionArraysEqual(
            self, repn.A.todense(), np.array([[-1, -2, -4], [5, 6, 8 * m.data[2]]])
        )
        assertExpressionListsEqual(
            self, repn.rhs, [-(10 * m.data[1] ** 2 - m.more_data), 20]
        )
        self.assertEqual(repn.rows, [(m.c, 1), (m.d, 1)])
        self.assertEqual(repn.columns, [m.x, m.y[1], m.y[3]])

    def test_parameterized_linear_model_row_col_order(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])
        m.data = Var([1, 2])
        m.more_data = Var()
        m.c = Constraint(expr=m.x + 2 * m.data[1] * m.data[2] * m.y[1] >= 3)
        m.d = Constraint(expr=m.y[1] + 4 * m.y[3] <= 5 * m.more_data)

        repn = ParameterizedLinearStandardFormCompiler().write(
            m,
            wrt=[m.data, m.more_data],
            column_order=[m.y[3], m.y[2], m.x, m.y[1]],
            row_order=[m.d, m.c],
        )

        self.assertTrue(np.all(repn.c == np.array([0, 0, 0])))
        assertExpressionArraysEqual(
            self,
            repn.A.todense(),
            np.array(
                [
                    [4, 0, 1],
                    [
                        0,
                        -1,
                        NegationExpression(
                            (
                                ProductExpression(
                                    [MonomialTermExpression([2, m.data[1]]), m.data[2]]
                                ),
                            )
                        ),
                    ],
                ]
            ),
        )
        assertExpressionListsEqual(self, repn.rhs, np.array([5 * m.more_data, -3]))
        self.assertEqual(repn.rows, [(m.d, 1), (m.c, -1)])
        self.assertEqual(repn.columns, [m.y[3], m.x, m.y[1]])

    def make_model(self, do_not_flip_c=False):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([0, 1, 3], bounds=lambda m, i: (-1 * (i % 2) * 5, 10 - 12 * (i // 2)))
        m.data = Var([1, 2])
        m.more_data = Var()
        if do_not_flip_c:
            # [ESJ: 06/24]: I should have done this sooner, but if you write c
            # this way, it gets interpreted as a >= constraint, which matches
            # the standard_form tests and makes life much easier. Unforuntately
            # I wrote a lot of tests before I thought of this, so I'm leaving in
            # both variations for the moment.
            m.c = Constraint(
                expr=m.data[1] ** 2 * m.x + 2 * m.y[1] - 3 * m.more_data >= 0
            )
        else:
            m.c = Constraint(expr=m.data[1] ** 2 * m.x + 2 * m.y[1] >= 3 * m.more_data)
        m.d = Constraint(expr=m.y[1] + 4 * m.y[3] <= 5 + m.data[2])
        m.e = Constraint(expr=inequality(-2, m.y[0] + 1 + 6 * m.y[1], 7))
        m.f = Constraint(expr=m.x + (m.data[2] + m.data[1] ** 3) * m.y[0] + 2 == 10)
        m.o = Objective([1, 3], rule=lambda m, i: m.x + i * 5 * m.more_data * m.y[i])
        m.o[1].sense = maximize

        return m

    def test_nonnegative_vars(self):
        m = self.make_model()
        col_order = [m.x, m.y[0], m.y[1], m.y[3]]
        repn = ParameterizedLinearStandardFormCompiler().write(
            m, wrt=[m.data, m.more_data], nonnegative_vars=True, column_order=col_order
        )

        # m.c comes back opposite how it does in test_standard_form, but that's
        # not unexpected.
        self.assertEqual(
            repn.rows, [(m.c, 1), (m.d, 1), (m.e, 1), (m.e, -1), (m.f, 1), (m.f, -1)]
        )
        self.assertEqual(
            list(map(str, repn.x)),
            ['_neg_0', '_pos_0', 'y[0]', '_neg_2', '_pos_2', '_neg_3'],
        )
        ref = np.array(
            [
                [
                    NegationExpression((ProductExpression((-1, m.data[1] ** 2)),)),
                    ProductExpression((-1, m.data[1] ** 2)),
                    0,
                    2,
                    -2,
                    0,
                ],
                [0, 0, 0, -1, 1, -4],
                [0, 0, 1, -6, 6, 0],
                [0, 0, -1, 6, -6, 0],
                [-1, 1, m.data[2] + m.data[1] ** 3, 0, 0, 0],
                [1, -1, -(m.data[2] + m.data[1] ** 3), 0, 0, 0],
            ]
        )
        assertExpressionArraysEqual(self, repn.A.todense(), ref)
        assertExpressionListsEqual(
            self,
            repn.b,
            [
                -3 * m.more_data,
                NegationExpression((ProductExpression((-1, 5 + m.data[2])),)),
                6,
                3,
                8,
                -8,
            ],
        )

        c_ref = np.array(
            [
                [1, -1, 0, 5 * m.more_data, -5 * m.more_data, 0],
                [-1, 1, 0, 0, 0, -15 * m.more_data],
            ]
        )
        assertExpressionArraysEqual(self, repn.c.todense(), c_ref)

    def test_slack_form(self):
        m = self.make_model()
        col_order = [m.x, m.y[0], m.y[1], m.y[3]]
        repn = ParameterizedLinearStandardFormCompiler().write(
            m, wrt=[m.data, m.more_data], slack_form=True, column_order=col_order
        )

        self.assertEqual(repn.rows, [(m.c, 1), (m.d, 1), (m.e, 1), (m.f, 1)])
        self.assertEqual(
            list(map(str, repn.x)),
            ['x', 'y[0]', 'y[1]', 'y[3]', '_slack_0', '_slack_1', '_slack_2'],
        )
        # m.c is flipped again, so the bounds on _slack_0 are flipped
        self.assertEqual(
            list(v.bounds for v in repn.x),
            [(None, None), (0, 10), (-5, 10), (-5, -2), (0, None), (0, None), (-9, 0)],
        )
        ref = np.array(
            [
                [ProductExpression((-1, m.data[1] ** 2)), 0, -2, 0, 1, 0, 0],
                [0, 0, 1, 4, 0, 1, 0],
                [0, 1, 6, 0, 0, 0, 1],
                [1, m.data[2] + m.data[1] ** 3, 0, 0, 0, 0, 0],
            ]
        )
        assertExpressionArraysEqual(self, repn.A.todense(), ref)
        assertExpressionListsEqual(
            self,
            repn.b,
            np.array(
                [
                    -3 * m.more_data,
                    NegationExpression((ProductExpression((-1, 5 + m.data[2])),)),
                    -3,
                    8,
                ]
            ),
        )
        c_ref = np.array(
            [
                [-1, 0, -5 * m.more_data, 0, 0, 0, 0],
                [1, 0, 0, 15 * m.more_data, 0, 0, 0],
            ]
        )
        assertExpressionArraysEqual(self, repn.c.todense(), c_ref)

    def test_mixed_form(self):
        m = self.make_model()
        col_order = [m.x, m.y[0], m.y[1], m.y[3]]
        repn = ParameterizedLinearStandardFormCompiler().write(
            m, wrt=[m.data, m.more_data], mixed_form=True, column_order=col_order
        )

        # m.c gets is opposite again
        self.assertEqual(repn.rows, [(m.c, 1), (m.d, 1), (m.e, 1), (m.e, -1), (m.f, 0)])
        self.assertEqual(list(map(str, repn.x)), ['x', 'y[0]', 'y[1]', 'y[3]'])
        self.assertEqual(
            list(v.bounds for v in repn.x), [(None, None), (0, 10), (-5, 10), (-5, -2)]
        )
        ref = np.array(
            [
                [ProductExpression((-1, m.data[1] ** 2)), 0, -2, 0],
                [0, 0, 1, 4],
                [0, 1, 6, 0],
                [0, 1, 6, 0],
                [1, m.data[2] + m.data[1] ** 3, 0, 0],
            ]
        )
        assertExpressionArraysEqual(self, repn.A.todense(), ref)
        assertExpressionListsEqual(
            self,
            repn.b,
            np.array(
                [
                    -3 * m.more_data,
                    NegationExpression((ProductExpression((-1, 5 + m.data[2])),)),
                    6,
                    -3,
                    8,
                ]
            ),
        )
        ref_c = np.array([[-1, 0, -5 * m.more_data, 0], [1, 0, 0, 15 * m.more_data]])
        assertExpressionArraysEqual(self, repn.c.todense(), ref_c)

    def test_slack_form_nonnegative_vars(self):
        m = self.make_model(do_not_flip_c=True)
        col_order = [m.x, m.y[0], m.y[1], m.y[3]]
        repn = ParameterizedLinearStandardFormCompiler().write(
            m,
            wrt=[m.data, m.more_data],
            slack_form=True,
            nonnegative_vars=True,
            column_order=col_order,
        )

        self.assertEqual(repn.rows, [(m.c, 1), (m.d, 1), (m.e, 1), (m.f, 1)])
        self.assertEqual(
            list(map(str, repn.x)),
            [
                '_neg_0',
                '_pos_0',
                'y[0]',
                '_neg_2',
                '_pos_2',
                '_neg_3',
                '_neg_4',
                '_slack_1',
                '_neg_6',
            ],
        )
        self.assertEqual(
            list(v.bounds for v in repn.x),
            [
                (0, None),
                (0, None),
                (0, 10),
                (0, 5),
                (0, 10),
                (2, 5),
                (0, None),
                (0, None),
                (0, 9),
            ],
        )
        ref = np.array(
            [
                [-m.data[1] ** 2, m.data[1] ** 2, 0, -2, 2, 0, -1, 0, 0],
                [0, 0, 0, -1, 1, -4, 0, 1, 0],
                [0, 0, 1, -6, 6, 0, 0, 0, -1],
                [-1, 1, m.data[2] + m.data[1] ** 3, 0, 0, 0, 0, 0, 0],
            ]
        )
        assertExpressionArraysEqual(self, repn.A.todense(), ref)
        assertExpressionListsEqual(
            self,
            repn.b,
            np.array(
                [
                    3 * m.more_data,
                    NegationExpression((ProductExpression((-1, 5 + m.data[2])),)),
                    -3,
                    8,
                ]
            ),
        )
        c_ref = np.array(
            [
                [1, -1, 0, 5 * m.more_data, -5 * m.more_data, 0, 0, 0, 0],
                [-1, 1, 0, 0, 0, -15 * m.more_data, 0, 0, 0],
            ]
        )
        assertExpressionArraysEqual(self, repn.c.todense(), c_ref)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import attempt_import

np, numpy_available = attempt_import(
    'numpy', 'Interior point requires numpy', minimum_version='1.13.0'
)
scipy, scipy_available = attempt_import('scipy', 'Interior point requires scipy')
mumps, mumps_available = attempt_import('mumps')
if not (numpy_available and scipy_available):
    raise unittest.SkipTest('Interior point tests require numpy and scipy')
from scipy.sparse import coo_matrix
import pyomo.contrib.interior_point as ip

if mumps_available:
    from pyomo.contrib.interior_point.linalg.mumps_interface import MumpsInterface
from pyomo.contrib.pynumero.linalg.base import LinearSolverStatus


@unittest.skipIf(not mumps_available, 'mumps is not available')
class TestReallocation(unittest.TestCase):
    def test_reallocate_memory_mumps(self):
        # Create a tri-diagonal matrix with small entries on the diagonal
        n = 10000
        small_val = 1e-7
        big_val = 1e2
        irn = []
        jcn = []
        ent = []
        for i in range(n - 1):
            irn.extend([i + 1, i, i])
            jcn.extend([i, i, i + 1])
            ent.extend([big_val, small_val, big_val])
        irn.append(n - 1)
        jcn.append(n - 1)
        ent.append(small_val)
        irn = np.array(irn)
        jcn = np.array(jcn)
        ent = np.array(ent)

        matrix = coo_matrix((ent, (irn, jcn)), shape=(n, n))

        linear_solver = MumpsInterface()
        linear_solver.do_symbolic_factorization(matrix)

        predicted = linear_solver.get_infog(16)

        # We predict that factorization will take 2 MB
        self.assertEqual(predicted, 2)

        # Explicitly set maximum memory to less than the predicted
        # requirement.
        linear_solver.set_icntl(23, 1)

        res = linear_solver.do_numeric_factorization(matrix, raise_on_error=False)
        self.assertEqual(res.status, LinearSolverStatus.not_enough_memory)

        linear_solver.do_symbolic_factorization(matrix)

        factor = 2
        linear_solver.increase_memory_allocation(factor)

        res = linear_solver.do_numeric_factorization(matrix)
        self.assertEqual(res.status, LinearSolverStatus.successful)

        # Expected memory allocation (MB)
        self.assertEqual(linear_solver._prev_allocation, 2 * predicted)

        actual = linear_solver.get_infog(18)

        # Sanity checks:
        # Make sure actual memory usage is greater than initial guess
        self.assertTrue(predicted < actual)
        # Make sure memory allocation is at least as much as was used
        self.assertTrue(actual <= linear_solver._prev_allocation)


if __name__ == '__main__':
    test_realloc = TestReallocation()
    test_realloc.test_reallocate_memory_mumps()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import pyomo.common.unittest as unittest
from pyomo.common.dependencies import numpy as np, numpy_available
from pyomo.common.dependencies import scipy_available, attempt_import

import pyomo.environ as pyo

# Use attempt_import here due to unguarded NumPy import in this file
nlp = attempt_import("pyomo.contrib.pynumero.interfaces.pyomo_nlp")[0]
import pyomo.contrib.sensitivity_toolbox.pynumero as pnsens
from pyomo.contrib.pynumero.asl import AmplInterface

if not scipy_available or not numpy_available:
    raise unittest.SkipTest("scipy or numpy is not available")

if not AmplInterface.available():
    raise unittest.SkipTest("Pynumero needs the ASL extension to run NLP tests")


class TestSeriesData(unittest.TestCase):
    def test_dsdp_dfdp_pyomo(self):
        m = pyo.ConcreteModel()
        m.x1 = pyo.Var(initialize=200)
        m.x2 = pyo.Var(initialize=5)
        m.p1 = pyo.Var(initialize=10)
        m.p2 = pyo.Var(initialize=5)
        m.obj = pyo.Objective(
            expr=m.x1 * m.p1 + m.x2 * m.x2 * m.p2 + m.p1 * m.p2, sense=pyo.minimize
        )
        m.c1 = pyo.Constraint(expr=m.x1 == 2 * m.p1**2)
        m.c2 = pyo.Constraint(expr=m.x2 == m.p2)
        theta = [m.p1, m.p2]

        dsdp, dfdp, rmap, cmap = pnsens.get_dsdp_dfdp(m, theta)

        # Since x1 = p1
        np.testing.assert_almost_equal(dsdp[rmap[m.x1], cmap[m.p1]], 40.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.x1], cmap[m.p2]], 0.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.x2], cmap[m.p1]], 0.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.x2], cmap[m.p2]], 1.0)

        # if x1 = 2 * p1 and x2 = p2 then
        #   df/dp1 = 6 p1**2 + p2 = 45.0
        #   df/dp2 = 3 p2 + p1 = 85.0
        np.testing.assert_almost_equal(dfdp[0, cmap[m.p1]], 605.0)
        np.testing.assert_almost_equal(dfdp[0, cmap[m.p2]], 85.0)

    def test_dsdp_dfdp_pyomo_nlp(self):
        m = pyo.ConcreteModel()
        m.x1 = pyo.Var(initialize=200)
        m.x2 = pyo.Var(initialize=5)
        m.p1 = pyo.Var(initialize=10)
        m.p2 = pyo.Var(initialize=5)
        m.obj = pyo.Objective(
            expr=m.x1 * m.p1 + m.x2 * m.x2 * m.p2 + m.p1 * m.p2, sense=pyo.minimize
        )
        m.c1 = pyo.Constraint(expr=m.x1 == 2 * m.p1**2)
        m.c2 = pyo.Constraint(expr=m.x2 == m.p2)
        theta = [m.p1, m.p2]

        m2 = nlp.PyomoNLP(m)
        dsdp, dfdp, rmap, cmap = pnsens.get_dsdp_dfdp(m2, theta)

        # Since x1 = p1
        assert dsdp.shape == (4, 2)
        np.testing.assert_almost_equal(dsdp[rmap[m.x1], cmap[m.p1]], 40.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.x1], cmap[m.p2]], 0.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.x2], cmap[m.p1]], 0.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.x2], cmap[m.p2]], 1.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.p1], cmap[m.p2]], 0.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.p2], cmap[m.p2]], 1.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.p1], cmap[m.p1]], 1.0)
        np.testing.assert_almost_equal(dsdp[rmap[m.p2], cmap[m.p1]], 0.0)

        # if x1 = 2 * p1 and x2 = p2 then
        #   df/dp1 = 6 p1**2 + p2 = 45.0
        #   df/dp2 = 3 p2 + p1 = 85.0
        assert dfdp.shape == (1, 2)
        np.testing.assert_almost_equal(dfdp[0, cmap[m.p1]], 605.0)
        np.testing.assert_almost_equal(dfdp[0, cmap[m.p2]], 85.0)

    def test_dydp_pyomo(self):
        m = pyo.ConcreteModel()
        m.x1 = pyo.Var(initialize=200)
        m.x2 = pyo.Var(initialize=5)
        m.p1 = pyo.Var(initialize=10)
        m.p2 = pyo.Var(initialize=5)
        m.obj = pyo.Objective(
            expr=m.x1 * m.p1 + m.x2 * m.x2 * m.p2 + m.p1 * m.p2, sense=pyo.minimize
        )
        m.c1 = pyo.Constraint(expr=m.x1 == 2 * m.p1**2)
        m.c2 = pyo.Constraint(expr=m.x2 == m.p2)

        m.e1 = pyo.Expression(expr=2 * m.p1**2)
        m.e2 = pyo.Expression(expr=m.p2)
        m.e3 = pyo.Expression(expr=m.x1 * m.p1 + m.x2 * m.x2 * m.p2 + m.p1 * m.p2)

        theta = [m.p1, m.p2]
        dsdp, dfdp, rmap, cmap = pnsens.get_dsdp_dfdp(m, theta)

        dydp, rmap = pnsens.get_dydp([m.e1, m.x1, m.e2, m.x2, m.e3], dsdp, rmap, cmap)

        assert dydp.shape == (5, 2)
        np.testing.assert_almost_equal(dydp[rmap[m.x1], cmap[m.p1]], 40.0)
        np.testing.assert_almost_equal(dydp[rmap[m.e1], cmap[m.p1]], 40.0)
        np.testing.assert_almost_equal(dydp[rmap[m.x1], cmap[m.p2]], 0.0)
        np.testing.assert_almost_equal(dydp[rmap[m.e1], cmap[m.p2]], 0.0)
        np.testing.assert_almost_equal(dydp[rmap[m.x2], cmap[m.p1]], 0.0)
        np.testing.assert_almost_equal(dydp[rmap[m.e2], cmap[m.p1]], 0.0)
        np.testing.assert_almost_equal(dydp[rmap[m.x2], cmap[m.p2]], 1.0)
        np.testing.assert_almost_equal(dydp[rmap[m.e2], cmap[m.p2]], 1.0)
        np.testing.assert_almost_equal(dydp[rmap[m.e3], cmap[m.p1]], 605.0)
        np.testing.assert_almost_equal(dydp[rmap[m.e3], cmap[m.p2]], 85.0)

        # make sure the rows are in the order of y_list
        assert rmap[m.e1] == 0
        assert rmap[m.x1] == 1
        assert rmap[m.e2] == 2
        assert rmap[m.x2] == 3
        assert rmap[m.e3] == 4



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import pyomo.contrib.mpc as mpc
from pyomo.contrib.mpc.examples.cstr.model import (
    create_instance,
    _plot_time_indexed_variables,
    _step_time_indexed_variables,
)


def get_input_sequence():
    input_sequence = mpc.TimeSeriesData(
        {"flow_in[*]": [0.1, 1.0, 0.5, 1.3, 1.0, 0.3]}, [0.0, 2.0, 4.0, 6.0, 8.0, 15.0]
    )
    return mpc.data.convert.series_to_interval(input_sequence)


def run_cstr_openloop(
    inputs, model_horizon=1.0, ntfe=10, simulation_steps=15, tee=False
):
    m = create_instance(horizon=model_horizon, ntfe=ntfe)
    dynamic_interface = mpc.DynamicModelInterface(m, m.time)

    sim_t0 = 0.0

    #
    # Initialize data structure to hold results of "rolling horizon"
    # simulation.
    #
    sim_data = dynamic_interface.get_data_at_time([sim_t0])

    solver = pyo.SolverFactory("ipopt")
    non_initial_model_time = list(m.time)[1:]
    for i in range(simulation_steps):
        # The starting point of this part of the simulation
        # in "real" time (rather than the model's time set)
        sim_t0 = i * model_horizon

        sim_time = [sim_t0 + t for t in m.time]
        new_inputs = mpc.data.convert.interval_to_series(inputs, time_points=sim_time)
        new_inputs.shift_time_points(m.time.first() - sim_t0)
        dynamic_interface.load_data(new_inputs, tolerance=1e-6)

        #
        # Solve square model to simulate
        #
        res = solver.solve(m, tee=tee)
        pyo.assert_optimal_termination(res)

        #
        # Extract data from simulated model
        #
        m_data = dynamic_interface.get_data_at_time(non_initial_model_time)
        m_data.shift_time_points(sim_t0 - m.time.first())
        sim_data.concatenate(m_data)

        #
        # Re-initialize (initial conditions and variable values)
        #
        # The default is to load this ScalarData at all points in the
        # model's time set.
        tf_data = dynamic_interface.get_data_at_time(m.time.last())
        dynamic_interface.load_data(tf_data)

    return m, sim_data


def main():
    input_sequence = get_input_sequence()
    m, sim_data = run_cstr_openloop(input_sequence, tee=False)
    _plot_time_indexed_variables(sim_data, [m.conc[:, "A"], m.conc[:, "B"]], show=True)
    _step_time_indexed_variables(sim_data, [m.flow_in[:]], show=True)


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.core.base import ConcreteModel, Var, Constraint, Objective
from pyomo.common.dependencies import attempt_import

np, numpy_available = attempt_import(
    'numpy', 'Interior point requires numpy', minimum_version='1.13.0'
)
scipy, scipy_available = attempt_import('scipy', 'Interior point requires scipy')
mumps, mumps_available = attempt_import('mumps', 'Interior point requires mumps')

if not (numpy_available and scipy_available):
    raise unittest.SkipTest('Interior point tests require numpy and scipy')

from pyomo.contrib.pynumero.asl import AmplInterface

asl_available = AmplInterface.available()
if not asl_available:
    raise unittest.SkipTest('Regularization tests require ASL')
from pyomo.contrib.interior_point.interface import InteriorPointInterface
from pyomo.contrib.interior_point.interior_point import InteriorPointSolver
from pyomo.contrib.interior_point.linalg.mumps_interface import MumpsInterface


def make_model_tri(n, small_val=1e-7, big_val=1e2):
    m = ConcreteModel()
    m.x = Var(range(n), initialize=0.5)

    def c_rule(m, i):
        return big_val * m.x[i - 1] + small_val * m.x[i] + big_val * m.x[i + 1] == 1

    m.c = Constraint(range(1, n - 1), rule=c_rule)

    m.obj = Objective(expr=small_val * sum((m.x[i] - 1) ** 2 for i in range(n)))

    return m


class TestReallocation(unittest.TestCase):
    def _test_ip_with_reallocation(self, linear_solver, interface):
        ip_solver = InteriorPointSolver(
            linear_solver,
            max_reallocation_iterations=3,
            reallocation_factor=1.1,
            # The small factor is to ensure that multiple iterations of
            # reallocation are performed. The bug in the previous
            # implementation only occurred if 2+ reallocation iterations
            # were needed (max_reallocation_iterations >= 3).
            max_iter=1,
        )
        ip_solver.set_interface(interface)

        ip_solver.solve(interface)

        return ip_solver

    @unittest.skipIf(not mumps_available, 'Mumps is not available')
    def test_mumps(self):
        n = 20000
        m = make_model_tri(n, small_val=1e-7)
        interface = InteriorPointInterface(m)
        linear_solver = MumpsInterface()
        # Default memory "buffer" factor: 20
        linear_solver.set_icntl(14, 20)

        kkt = interface.evaluate_primal_dual_kkt_matrix()
        res = linear_solver.do_symbolic_factorization(kkt)
        predicted = linear_solver.get_infog(16)

        linear_solver.set_icntl(23, 8)

        self._test_ip_with_reallocation(linear_solver, interface)
        # In Mumps 5.6.2 (and likely previous versions), ICNTL(23)=0
        # corresponds to "use default increase factor over prediction".
        actual = linear_solver.get_icntl(23)
        percent_increase = linear_solver.get_icntl(14)
        increase_factor = 1.0 + percent_increase / 100.0

        if actual == 0:
            actual = increase_factor * predicted

        # As of Mumps 5.6.2, predicted == 9, which is lower than the
        # default actual of 10.8
        # self.assertTrue(predicted == 12 or predicted == 11)
        self.assertTrue(actual >= predicted)
        # NOTE: This test will break if Mumps (or your Mumps version)
        # gets more conservative at estimating memory requirement,
        # or if the numeric factorization gets more efficient.


if __name__ == '__main__':
    #
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

import pyomo.environ as pyo
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.series_data import TimeSeriesData


class TestSeriesData(unittest.TestCase):
    def _make_model(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[0.0, 0.1, 0.2])
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(m.time, m.comp, initialize=1.0)
        return m

    def test_construct_and_get_data(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)

        processed_data_dict = {
            pyo.ComponentUID(key): val for key, val in data_dict.items()
        }
        self.assertEqual(data.get_data(), processed_data_dict)

    def test_construct_exception(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4]}
        msg = "must have same length"
        with self.assertRaisesRegex(ValueError, msg):
            data = TimeSeriesData(data_dict, m.time)

        data_dict = {m.var[:, "A"]: [1, 2], m.var[:, "B"]: [2, 4]}
        with self.assertRaisesRegex(ValueError, msg):
            # series don't have same lengths as time
            data = TimeSeriesData(data_dict, m.time)

        msg = "not sorted"
        with self.assertRaisesRegex(ValueError, msg):
            # Time list has right number of points, but is not sorted
            # increasing.
            data = TimeSeriesData(data_dict, [0, -1])

    def test_get_time_points(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)
        self.assertEqual(data.get_time_points(), list(m.time))

        new_time_list = [3, 4, 5]
        data = TimeSeriesData(data_dict, new_time_list)
        self.assertEqual(data.get_time_points(), new_time_list)

    def test_get_data_at_time_indices(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)
        new_data = data.get_data_at_time_indices(1)
        self.assertEqual(ScalarData({m.var[:, "A"]: 2, m.var[:, "B"]: 4}), new_data)

        new_data = data.get_data_at_time_indices([1])
        t1 = m.time.at(2)  # Sets are indexed starting from 1...
        self.assertEqual(
            TimeSeriesData({m.var[:, "A"]: [2], m.var[:, "B"]: [4]}, [t1]), new_data
        )

        new_t = [m.time.at(1), m.time.at(3)]
        new_data = data.get_data_at_time_indices([0, 2])
        self.assertEqual(
            TimeSeriesData({m.var[:, "A"]: [1, 3], m.var[:, "B"]: [2, 6]}, new_t),
            new_data,
        )

    def test_get_data_at_time(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)
        new_data = data.get_data_at_time(0.1)
        self.assertEqual(ScalarData({m.var[:, "A"]: 2, m.var[:, "B"]: 4}), new_data)

        t1 = 0.1
        new_data = data.get_data_at_time([t1])
        self.assertEqual(
            TimeSeriesData({m.var[:, "A"]: [2], m.var[:, "B"]: [4]}, [t1]), new_data
        )

        new_t = [0.0, 0.2]
        new_data = data.get_data_at_time(new_t)
        self.assertEqual(
            TimeSeriesData({m.var[:, "A"]: [1, 3], m.var[:, "B"]: [2, 6]}, new_t),
            new_data,
        )

    def test_get_data_at_time_with_tolerance(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)

        # Test an invalid time value. A tolerance of None gives us
        # the closest index
        new_data = data.get_data_at_time(-0.1, tolerance=None)
        self.assertEqual(ScalarData({m.var[:, "A"]: 1, m.var[:, "B"]: 2}), new_data)

        # Test a value that is only valid within tolerance
        new_data = data.get_data_at_time(-0.0001, tolerance=1e-3)
        self.assertEqual(ScalarData({m.var[:, "A"]: 1, m.var[:, "B"]: 2}), new_data)

        # The default is to raise an error in the case of any discrepancy.
        msg = "Time point.*is invalid"
        with self.assertRaisesRegex(RuntimeError, msg):
            new_data = data.get_data_at_time(-0.0001)

        # Test a value that is invalid within tolerance
        msg = "Time point.*is invalid"
        with self.assertRaisesRegex(RuntimeError, msg):
            new_data = data.get_data_at_time(-0.01, tolerance=1e-3)

    def test_get_data_interpolate(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)
        new_data = data.get_interpolated_data(0.05)
        self.assertEqual(ScalarData({m.var[:, "A"]: 1.5, m.var[:, "B"]: 3}), new_data)

        t1 = 0.05
        new_data = data.get_interpolated_data([t1])
        self.assertEqual(
            TimeSeriesData({m.var[:, "A"]: [1.5], m.var[:, "B"]: [3]}, [t1]), new_data
        )

        new_t = [0.05, 0.15]
        new_data = data.get_interpolated_data(new_t)
        self.assertEqual(
            TimeSeriesData({m.var[:, "A"]: [1.5, 2.5], m.var[:, "B"]: [3, 5]}, new_t),
            new_data,
        )

    def test_get_data_interpolate_range_check(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)
        msg = "Requesting interpolation outside data range."
        with self.assertRaisesRegex(RuntimeError, msg):
            new_data = data.get_interpolated_data(0.2 + 1e-6)
        with self.assertRaisesRegex(RuntimeError, msg):
            new_data = data.get_interpolated_data(0.0 - 1e-6)
        new_data = data.get_interpolated_data(0.2 + 1e-6, tolerance=1e-5)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "A"]), 3, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "B"]), 6, 4)

        t1 = 0.2 + 1e-6
        with self.assertRaisesRegex(RuntimeError, msg):
            new_data = data.get_interpolated_data([t1])
        new_data = data.get_interpolated_data([t1], tolerance=1e-5)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "A"])[0], 3, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "B"])[0], 6, 4)

        new_t = [0.0 - 1e-6, 0.2 + 1e-6]
        with self.assertRaisesRegex(RuntimeError, msg):
            new_data = data.get_interpolated_data(new_t)
        new_data = data.get_interpolated_data(new_t, tolerance=1e-5)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "A"])[0], 1, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "B"])[0], 2, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "A"])[1], 3, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "B"])[1], 6, 4)

        # check that the exact endpoints don't raise an exception with 0 tol
        new_t = [0.0, 0.2]
        new_data = data.get_interpolated_data(new_t, tolerance=0.0)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "A"])[0], 1, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "B"])[0], 2, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "A"])[1], 3, 4)
        self.assertAlmostEqual(new_data.get_data_from_key(m.var[:, "B"])[1], 6, 4)

    def test_to_serializable(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time).to_serializable()

        pred_json_obj = (
            {str(pyo.ComponentUID(key)): val for key, val in data_dict.items()},
            list(m.time),
        )
        self.assertEqual(data, pred_json_obj)

        # Test attributes of the TimeSeriesTuple namedtuple
        self.assertEqual(data.time, list(m.time))
        self.assertEqual(
            data.data,
            {str(pyo.ComponentUID(key)): val for key, val in data_dict.items()},
        )

    def test_concatenate(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        time1 = [t for t in m.time]
        data1 = TimeSeriesData(data_dict, time1)

        time2 = [t + 1.0 for t in m.time]
        data2 = TimeSeriesData(data_dict, time2)

        data1.concatenate(data2)
        pred_time = time1 + time2
        pred_data = {
            m.var[:, "A"]: [1, 2, 3, 1, 2, 3],
            m.var[:, "B"]: [2, 4, 6, 2, 4, 6],
        }
        # Note that data1 has been modified in place
        self.assertEqual(TimeSeriesData(pred_data, pred_time), data1)

    def test_concatenate_exception(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        time1 = [t for t in m.time]
        data1 = TimeSeriesData(data_dict, time1)

        msg = "Initial time point.*is not greater"
        with self.assertRaisesRegex(ValueError, msg):
            data1.concatenate(data1)

    def test_shift_time_points(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)

        offset = 1.0
        data.shift_time_points(offset)
        self.assertEqual(data.get_time_points(), [t + offset for t in m.time])

    def test_extract_variables(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)

        new_data = data.extract_variables([m.var[:, "A"]])
        self.assertEqual(new_data, TimeSeriesData({m.var[:, "A"]: [1, 2, 3]}, m.time))

    def test_shift_then_get_data(self):
        m = self._make_model()
        data_dict = {m.var[:, "A"]: [1, 2, 3], m.var[:, "B"]: [2, 4, 6]}
        data = TimeSeriesData(data_dict, m.time)

        offset = 0.1
        data.shift_time_points(offset)
        self.assertEqual(data.get_time_points(), [t + offset for t in m.time])

        # A time point of zero is no longer valid
        msg = "Time point.*is invalid"
        with self.assertRaisesRegex(RuntimeError, msg):
            t0_data = data.get_data_at_time(0.0, tolerance=1e-3)

        t1_data = data.get_data_at_time(0.1)
        self.assertEqual(t1_data, ScalarData({m.var[:, "A"]: 1, m.var[:, "B"]: 2}))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import glob
import shutil
from os.path import join, basename, dirname, isfile

from pyomo.opt.base.solvers import _extract_version


class MockMIP:
    """Methods used to create a mock MIP solver used for testing"""

    def __init__(self, mockdir):
        self.mock_subdir = mockdir

    def create_command_line(self, executable, problem_files):
        self._mock_problem = basename(problem_files[0]).split('.')[0]
        self._mock_dir = dirname(problem_files[0])

    def _default_executable(self):
        return "mock"

    executable = _default_executable

    def version(self):
        return _extract_version('')

    def _execute_command(self, cmd):
        mock_basename = join(self._mock_dir, self.mock_subdir, self._mock_problem)
        if self._soln_file is not None:
            # prefer .sol over .soln
            for ext in ('sol', 'soln'):
                file = glob.glob(mock_basename + "*." + ext)
                if len(file):
                    if len(file) > 1:
                        raise RuntimeError("Multiple .%s files found" % ext)
                    shutil.copyfile(file[0], self._soln_file)
                    break
        for file in glob.glob(mock_basename + "*"):
            if file.split(".")[-1] != "out":
                shutil.copyfile(file, join(self._mock_dir, basename(file)))
        log = ""
        fname = mock_basename + ".out"
        if not isfile(fname):
            raise ValueError("Missing mock data file: " + fname)
        INPUT = open(mock_basename + ".out")
        for line in INPUT:
            log = log + line
        INPUT.close()
        return [0, log]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)
from pyomo.common.dependencies.scipy import sparse as spa

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

from ..external_grey_box import ExternalGreyBoxModel, ExternalGreyBoxBlock

# set of external models for testing
# basic model is a simple pipe sequence with nonlinear pressure drop
# Pin -> P1 -> P2 -> P3 -> Pout
#
# We will assume that we have an external model to compute
# the pressure drop in this sequence of pipes, where the dP
# is given by c*F^2
#
# There are several ways to format this.
# Model 1: Use the "external model" to compute the output pressure
#   no equalities, 1 output
#   u = [Pin, c, F]
#   o = [Pout]
#   h_eq(u) = {empty}
#   h_o(u) = [Pin - 4*c*F^2]
#
# Model 2: Same as model 1, but treat Pout as an input to be converged by the optimizer
#   1 equality, no outputs
#   u = [Pin, c, F, Pout]
#   o = {empty}
#   h_eq(u) = [Pout - (Pin - 4*c*F^2]
#   h_o(u) = {empty}
#
# Model 3: Use the "external model" to compute the output pressure and the pressure
#          at node 2 (e.g., maybe we have a measurement there we want to match)
#   no equalities, 2 outputs
#   u = [Pin, c, F]
#   o = [P2, Pout]
#   h_eq(u) = {empty}
#   h_o(u) = [Pin - 2*c*F^2]
#            [Pin - 4*c*F^2]
#
# Model 4: Same as model 2, but treat P2, and Pout as an input to be converged by the optimizer
#   2 equality, no outputs
#   u = [Pin, c, F, P2, Pout]
#   o = {empty}
#   h_eq(u) = [P2 - (Pin - 2*c*F^2]
#             [Pout - (P2 - 2*c*F^2]
#   h_o(u) = {empty}


# Model 4: Same as model 2, but treat P2 as an input to be converged by the solver
#   u = [Pin, c, F, P2]
#   o = [Pout]
#   h_eq(u) = P2 - (Pin-2*c*F^2)]
#   h_o(u) = [Pin - 4*c*F^2] (or could also be [P2 - 2*c*F^2])
#
# Model 5: treat all "internal" variables as "inputs", equality and output equations
#   u = [Pin, c, F, P1, P2, P3]
#   o = [Pout]
#    h_eq(u) = [
#               P1 - (Pin - c*F^2);
#               P2 - (P1 - c*F^2);
#               P3 - (P2 - c*F^2);
#              ]
#   h_o(u) = [P3 - c*F^2] (or could also be [Pin - 4*c*F^2] or [P1 - 3*c*F^2] or [P2 - 2*c*F^2])
#
# Model 6: treat all variables as "inputs", equality only, and no output equations
#   u = [Pin, c, F, P1, P2, P3, Pout]
#   o = {empty}
#   h_eq(u) = [
#               P1 - (Pin - c*F^2);
#               P2 - (P1 - c*F^2);
#               P3 - (P2 - c*F^2);
#               Pout = (P3 - c*F^2);
#              ]
#   h_o(u) = {empty}
#
class PressureDropSingleOutput(ExternalGreyBoxModel):
    def __init__(self):
        self._input_names = ['Pin', 'c', 'F']
        self._input_values = np.zeros(3, dtype=np.float64)
        self._output_names = ['Pout']

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return []

    def output_names(self):
        return self._output_names

    def set_input_values(self, input_values):
        assert len(input_values) == 3
        np.copyto(self._input_values, input_values)

    def evaluate_outputs(self):
        Pin = self._input_values[0]
        c = self._input_values[1]
        F = self._input_values[2]
        Pout = Pin - 4 * c * F**2
        return np.asarray([Pout], dtype=np.float64)

    def evaluate_jacobian_outputs(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([0, 0, 0], dtype=np.int64)
        jcol = np.asarray([0, 1, 2], dtype=np.int64)
        nonzeros = np.asarray([1, -4 * F**2, -4 * c * 2 * F], dtype=np.float64)
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(1, 3))
        return jac


class PressureDropSingleOutputWithHessian(PressureDropSingleOutput):
    def __init__(self):
        super(PressureDropSingleOutputWithHessian, self).__init__()
        self._output_con_mult_values = np.zeros(1, dtype=np.float64)

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        np.copyto(self._output_con_mult_values, output_con_multiplier_values)

    def evaluate_hessian_outputs(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([2, 2], dtype=np.int64)
        jcol = np.asarray([1, 2], dtype=np.int64)
        data = self._output_con_mult_values[0] * np.asarray(
            [-8 * F, -8 * c], dtype=np.float64
        )
        hess = spa.coo_matrix((data, (irow, jcol)), shape=(3, 3))
        return hess


class PressureDropSingleEquality(ExternalGreyBoxModel):
    #   u = [Pin, c, F, Pout]
    #   o = {empty}
    #   h_eq(u) = [Pout - (Pin - 4*c*F^2]
    #   h_o(u) = {empty}
    def __init__(self):
        self._input_names = ['Pin', 'c', 'F', 'Pout']
        self._input_values = np.zeros(4, dtype=np.float64)
        self._equality_constraint_names = ['pdrop']

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return self._equality_constraint_names

    def set_input_values(self, input_values):
        assert len(input_values) == 4
        np.copyto(self._input_values, input_values)

    def evaluate_equality_constraints(self):
        Pin = self._input_values[0]
        c = self._input_values[1]
        F = self._input_values[2]
        Pout = self._input_values[3]
        return np.asarray([Pout - (Pin - 4 * c * F**2)], dtype=np.float64)

    def evaluate_jacobian_equality_constraints(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([0, 0, 0, 0], dtype=np.int64)
        jcol = np.asarray([0, 1, 2, 3], dtype=np.int64)
        nonzeros = np.asarray([-1, 4 * F**2, 4 * 2 * c * F, 1], dtype=np.float64)
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(1, 4))
        return jac


class PressureDropSingleEqualityWithHessian(PressureDropSingleEquality):
    #   u = [Pin, c, F, Pout]
    #   o = {empty}
    #   h_eq(u) = [Pout - (Pin - 4*c*F^2]
    #   h_o(u) = {empty}
    def __init__(self):
        super(PressureDropSingleEqualityWithHessian, self).__init__()
        self._eq_con_mult_values = np.zeros(1, dtype=np.float64)

    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        assert len(eq_con_multiplier_values) == 1
        np.copyto(self._eq_con_mult_values, eq_con_multiplier_values)

    def evaluate_hessian_equality_constraints(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([2, 2], dtype=np.int64)
        jcol = np.asarray([1, 2], dtype=np.int64)
        nonzeros = self._eq_con_mult_values[0] * np.asarray(
            [8 * F, 8 * c], dtype=np.float64
        )
        hess = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(4, 4))
        return hess


class PressureDropTwoOutputs(ExternalGreyBoxModel):
    #   u = [Pin, c, F]
    #   o = [P2, Pout]
    #   h_eq(u) = {empty}
    #   h_o(u) = [Pin - 2*c*F^2]
    #            [Pin - 4*c*F^2]
    def __init__(self):
        self._input_names = ['Pin', 'c', 'F']
        self._input_values = np.zeros(3, dtype=np.float64)
        self._output_names = ['P2', 'Pout']

    def input_names(self):
        return self._input_names

    def output_names(self):
        return self._output_names

    def set_input_values(self, input_values):
        assert len(input_values) == 3
        np.copyto(self._input_values, input_values)

    def evaluate_equality_constraints(self):
        raise NotImplementedError('This method should not be called for this model.')

    def evaluate_outputs(self):
        Pin = self._input_values[0]
        c = self._input_values[1]
        F = self._input_values[2]
        P2 = Pin - 2 * c * F**2
        Pout = Pin - 4 * c * F**2
        return np.asarray([P2, Pout], dtype=np.float64)

    def evaluate_jacobian_outputs(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([0, 0, 0, 1, 1, 1], dtype=np.int64)
        jcol = np.asarray([0, 1, 2, 0, 1, 2], dtype=np.int64)
        nonzeros = np.asarray(
            [1, -2 * F**2, -2 * c * 2 * F, 1, -4 * F**2, -4 * c * 2 * F],
            dtype=np.float64,
        )
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(2, 3))
        return jac


class PressureDropTwoOutputsWithHessian(PressureDropTwoOutputs):
    #   u = [Pin, c, F]
    #   o = [P2, Pout]
    #   h_eq(u) = {empty}
    #   h_o(u) = [Pin - 2*c*F^2]
    #            [Pin - 4*c*F^2]
    def __init__(self):
        super(PressureDropTwoOutputsWithHessian, self).__init__()
        self._output_con_mult_values = np.zeros(2, dtype=np.float64)

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        np.copyto(self._output_con_mult_values, output_con_multiplier_values)

    def evaluate_hessian_outputs(self):
        c = self._input_values[1]
        F = self._input_values[2]
        y1 = self._output_con_mult_values[0]
        y2 = self._output_con_mult_values[1]
        irow = np.asarray([2, 2], dtype=np.int64)
        jcol = np.asarray([1, 2], dtype=np.int64)
        nonzeros = np.asarray(
            [y1 * (-4 * F) + y2 * (-8 * F), y1 * (-4 * c) + y2 * (-8 * c)],
            dtype=np.float64,
        )
        hess = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(3, 3))
        return hess


class PressureDropTwoEqualities(ExternalGreyBoxModel):
    #   u = [Pin, c, F, P2, Pout]
    #   o = {empty}
    #   h_eq(u) = [P2 - (Pin - 2*c*F^2]
    #             [Pout - (P2 - 2*c*F^2]
    #   h_o(u) = {empty}
    def __init__(self):
        self._input_names = ['Pin', 'c', 'F', 'P2', 'Pout']
        self._input_values = np.zeros(5, dtype=np.float64)
        self._equality_constraint_names = ['pdrop2', 'pdropout']

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return self._equality_constraint_names

    def set_input_values(self, input_values):
        assert len(input_values) == 5
        np.copyto(self._input_values, input_values)

    def evaluate_equality_constraints(self):
        Pin = self._input_values[0]
        c = self._input_values[1]
        F = self._input_values[2]
        P2 = self._input_values[3]
        Pout = self._input_values[4]
        return np.asarray(
            [P2 - (Pin - 2 * c * F**2), Pout - (P2 - 2 * c * F**2)], dtype=np.float64
        )

    def evaluate_jacobian_equality_constraints(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([0, 0, 0, 0, 1, 1, 1, 1], dtype=np.int64)
        jcol = np.asarray([0, 1, 2, 3, 1, 2, 3, 4], dtype=np.int64)
        nonzeros = np.asarray(
            [-1, 2 * F**2, 2 * 2 * c * F, 1, 2 * F**2, 2 * 2 * c * F, -1, 1],
            dtype=np.float64,
        )
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(2, 5))
        return jac


class PressureDropTwoEqualitiesWithHessian(PressureDropTwoEqualities):
    #   u = [Pin, c, F, P2, Pout]
    #   o = {empty}
    #   h_eq(u) = [P2 - (Pin - 2*c*F^2]
    #             [Pout - (P2 - 2*c*F^2]
    #   h_o(u) = {empty}
    def __init__(self):
        super(PressureDropTwoEqualitiesWithHessian, self).__init__()
        self._eq_con_mult_values = np.zeros(2, dtype=np.float64)

    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        assert len(eq_con_multiplier_values) == 2
        np.copyto(self._eq_con_mult_values, eq_con_multiplier_values)

    def evaluate_hessian_equality_constraints(self):
        c = self._input_values[1]
        F = self._input_values[2]
        y1 = self._eq_con_mult_values[0]
        y2 = self._eq_con_mult_values[1]

        irow = np.asarray([2, 2], dtype=np.int64)
        jcol = np.asarray([1, 2], dtype=np.int64)
        nonzeros = np.asarray(
            [y1 * (4 * F) + y2 * (4 * F), y1 * (4 * c) + y2 * (4 * c)], dtype=np.float64
        )
        hess = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(5, 5))
        return hess


class PressureDropTwoEqualitiesTwoOutputs(ExternalGreyBoxModel):
    #   u = [Pin, c, F, P1, P3]
    #   o = {P2, Pout}
    #   h_eq(u) = [P1 - (Pin - c*F^2]
    #             [P3 - (P1 - 2*c*F^2]
    #   h_o(u) = [P1 - c*F^2]
    #            [Pin - 4*c*F^2]
    def __init__(self):
        self._input_names = ['Pin', 'c', 'F', 'P1', 'P3']
        self._input_values = np.zeros(5, dtype=np.float64)
        self._equality_constraint_names = ['pdrop1', 'pdrop3']
        self._output_names = ['P2', 'Pout']

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return self._equality_constraint_names

    def output_names(self):
        return self._output_names

    def set_input_values(self, input_values):
        assert len(input_values) == 5
        np.copyto(self._input_values, input_values)

    def evaluate_equality_constraints(self):
        Pin = self._input_values[0]
        c = self._input_values[1]
        F = self._input_values[2]
        P1 = self._input_values[3]
        P3 = self._input_values[4]
        return np.asarray(
            [P1 - (Pin - c * F**2), P3 - (P1 - 2 * c * F**2)], dtype=np.float64
        )

    def evaluate_outputs(self):
        Pin = self._input_values[0]
        c = self._input_values[1]
        F = self._input_values[2]
        P1 = self._input_values[3]
        return np.asarray([P1 - c * F**2, Pin - 4 * c * F**2], dtype=np.float64)

    def evaluate_jacobian_equality_constraints(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([0, 0, 0, 0, 1, 1, 1, 1], dtype=np.int64)
        jcol = np.asarray([0, 1, 2, 3, 1, 2, 3, 4], dtype=np.int64)
        nonzeros = np.asarray(
            [-1, F**2, 2 * c * F, 1, 2 * F**2, 4 * c * F, -1, 1], dtype=np.float64
        )
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(2, 5))
        return jac

    def evaluate_jacobian_outputs(self):
        c = self._input_values[1]
        F = self._input_values[2]
        irow = np.asarray([0, 0, 0, 1, 1, 1], dtype=np.int64)
        jcol = np.asarray([1, 2, 3, 0, 1, 2], dtype=np.int64)
        nonzeros = np.asarray(
            [-(F**2), -c * 2 * F, 1, 1, -4 * F**2, -4 * c * 2 * F], dtype=np.float64
        )
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(2, 5))
        return jac


class PressureDropTwoEqualitiesTwoOutputsWithHessian(
    PressureDropTwoEqualitiesTwoOutputs
):
    #   u = [Pin, c, F, P1, P3]
    #   o = {P2, Pout}
    #   h_eq(u) = [P1 - (Pin - c*F^2]
    #             [P3 - (P1 - 2*c*F^2]
    #   h_o(u) = [P1 - c*F^2]
    #            [Pin - 4*c*F^2]
    def __init__(self):
        super(PressureDropTwoEqualitiesTwoOutputsWithHessian, self).__init__()
        self._eq_con_mult_values = np.zeros(2, dtype=np.float64)
        self._output_con_mult_values = np.zeros(2, dtype=np.float64)

    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        assert len(eq_con_multiplier_values) == 2
        np.copyto(self._eq_con_mult_values, eq_con_multiplier_values)

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        assert len(output_con_multiplier_values) == 2
        np.copyto(self._output_con_mult_values, output_con_multiplier_values)

    def evaluate_hessian_equality_constraints(self):
        c = self._input_values[1]
        F = self._input_values[2]
        y1 = self._eq_con_mult_values[0]
        y2 = self._eq_con_mult_values[1]
        irow = np.asarray([2, 2], dtype=np.int64)
        jcol = np.asarray([1, 2], dtype=np.int64)
        nonzeros = np.asarray(
            [y1 * (2 * F) + y2 * (4 * F), y1 * (2 * c) + y2 * (4 * c)], dtype=np.float64
        )
        hess = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(5, 5))
        return hess

    def evaluate_hessian_outputs(self):
        c = self._input_values[1]
        F = self._input_values[2]
        y1 = self._output_con_mult_values[0]
        y2 = self._output_con_mult_values[1]
        irow = np.asarray([2, 2], dtype=np.int64)
        jcol = np.asarray([1, 2], dtype=np.int64)
        nonzeros = np.asarray(
            [y1 * (-2 * F) + y2 * (-8 * F), y1 * (-2 * c) + y2 * (-8 * c)],
            dtype=np.float64,
        )
        hess = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(5, 5))
        return hess


class PressureDropTwoEqualitiesTwoOutputsScaleBoth(PressureDropTwoEqualitiesTwoOutputs):
    def get_equality_constraint_scaling_factors(self):
        return np.asarray([3.1, 3.2], dtype=np.float64)

    def get_output_constraint_scaling_factors(self):
        return np.asarray([4.1, 4.2])


class PressureDropTwoEqualitiesTwoOutputsScaleEqualities(
    PressureDropTwoEqualitiesTwoOutputs
):
    def get_equality_constraint_scaling_factors(self):
        return np.asarray([3.1, 3.2], dtype=np.float64)


class PressureDropTwoEqualitiesTwoOutputsScaleOutputs(
    PressureDropTwoEqualitiesTwoOutputs
):
    def get_output_constraint_scaling_factors(self):
        return np.asarray([4.1, 4.2])


class OneOutput(ExternalGreyBoxModel):
    def __init__(self):
        self._input_names = ['u']
        self._output_names = ['o']
        self._u = None
        self._output_mult = None

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return []

    def output_names(self):
        return self._output_names

    def finalize_block_construction(self, pyomo_block):
        pyomo_block.inputs['u'].setlb(4)
        pyomo_block.inputs['u'].setub(10)
        pyomo_block.inputs['u'].set_value(1.0)
        pyomo_block.outputs['o'].set_value(1.0)

    def set_input_values(self, input_values):
        assert len(input_values) == 1
        self._u = input_values[0]

    def evaluate_outputs(self):
        return np.asarray([5 * self._u])

    def evaluate_jacobian_outputs(self):
        irow = np.asarray([0], dtype=np.int64)
        jcol = np.asarray([0], dtype=np.int64)
        nonzeros = np.asarray([5.0], dtype=np.float64)
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(1, 1))
        return jac

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        assert len(output_con_multiplier_values) == 1
        self._output_mult = output_con_multiplier_values[0]

    def evaluate_hessian_outputs(self):
        irow = np.asarray([], dtype=np.int64)
        jcol = np.asarray([], dtype=np.int64)
        data = np.asarray([], dtype=np.float64)
        hess = spa.coo_matrix((data, (irow, jcol)), shape=(1, 1))
        return hess


class OneOutputOneEquality(ExternalGreyBoxModel):
    def __init__(self):
        self._input_names = ['u']
        self._equality_constraint_names = ['u2_con']
        self._output_names = ['o']
        self._u = None
        self._output_mult = None
        self._equality_mult = None

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return self._equality_constraint_names

    def output_names(self):
        return self._output_names

    def finalize_block_construction(self, pyomo_block):
        pyomo_block.inputs['u'].set_value(1.0)
        pyomo_block.outputs['o'].set_value(1.0)

    def set_input_values(self, input_values):
        assert len(input_values) == 1
        self._u = input_values[0]

    def evaluate_equality_constraints(self):
        return np.asarray([self._u**2 - 1])

    def evaluate_outputs(self):
        return np.asarray([5 * self._u])

    def evaluate_jacobian_equality_constraints(self):
        irow = np.asarray([0], dtype=np.int64)
        jcol = np.asarray([0], dtype=np.int64)
        nonzeros = np.asarray([2 * self._u], dtype=np.float64)
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(1, 1))
        return jac

    def evaluate_jacobian_outputs(self):
        irow = np.asarray([0], dtype=np.int64)
        jcol = np.asarray([0], dtype=np.int64)
        nonzeros = np.asarray([5.0], dtype=np.float64)
        jac = spa.coo_matrix((nonzeros, (irow, jcol)), shape=(1, 1))
        return jac

    def set_equality_constraint_multipliers(self, equality_con_multiplier_values):
        assert len(equality_con_multiplier_values) == 1
        self._equality_mult = equality_con_multiplier_values[0]

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        assert len(output_con_multiplier_values) == 1
        self._output_mult = output_con_multiplier_values[0]

    def evaluate_hessian_equality_constraints(self):
        irow = np.asarray([0], dtype=np.int64)
        jcol = np.asarray([0], dtype=np.int64)
        data = np.asarray([self._equality_mult * 2.0], dtype=np.float64)
        hess = spa.coo_matrix((data, (irow, jcol)), shape=(1, 1))
        return hess

    def evaluate_hessian_outputs(self):
        irow = np.asarray([], dtype=np.int64)
        jcol = np.asarray([], dtype=np.int64)
        data = np.asarray([], dtype=np.float64)
        hess = spa.coo_matrix((data, (irow, jcol)), shape=(1, 1))
        return hess
